from ..attributetype import AttributeType
from ..exceptions import LDAPExtensionError
from ..controls import Control
from ..objectclass import ObjectClass
from ..rules import SyntaxRule, MatchingRule
from ..utils import get_obj_module
import six
import logging

logger = logging.getLogger(__name__)

_preregistered_elements = {}
_registered_mods = set()

_recognized_subclasses = (SyntaxRule, MatchingRule, Control)
_recognized_instances = (ObjectClass, AttributeType)


class MetaLaurelinTransiter(type):
    """Prepares schema elements and controls defined as attributes of infraclasses for registration"""
    def __new__(mcs, name, bases, dct):
        cls = type.__new__(mcs, name, bases, dct)
        modname = get_obj_module(cls)

        for attr, value in dct.items():
            if attr.startswith('_'):
                continue

            if isinstance(value, six.class_types) and issubclass(value, _recognized_subclasses):
                is_element = True
            elif isinstance(value, _recognized_instances):
                is_element = True
            else:
                is_element = False

            if is_element:
                if modname not in _preregistered_elements:
                    _preregistered_elements[modname] = []
                _preregistered_elements[modname].append(value)

        return cls


@six.add_metaclass(MetaLaurelinTransiter)
class LaurelinTransiter(object):
    """Base class for classes in extensions defining schema elements or controls"""
    pass


class LaurelinRegistrar(object):
    """The require() method on this class registers all schema and controls in a module.

    If this class is subclassed in an extension module (such as with LaurelinExtension), there is no need to pass an
    argument to the constructor. However, for extensions with many schema elements or controls, it may be desirable to
    break up these objects into multiple submodules that are imported on end-user request.

    In such a setup, an instance of LaurelinRegistrar would need to be created and exposed to the end-user for each
    submodule so that they can call ``.require()`` on it. This constructor would need to be passed ``__name__``.

    Note that if this is applied to a package, require() will function on all imported submodules of the package.
    """

    def __init__(self, modname=None):
        if modname:
            self.modname = modname
        else:
            self.modname = get_obj_module(self.__class__)
            if self.modname == __name__:
                raise LDAPExtensionError('You must pass __name__ to LaurelinRegistrar()')

    def require(self):
        """Register schema elements and controls defined in any LaurelinTransiter subclass in the same module as the
        calling class. If the calling class is defined in an __init__.py, also register schema/controls for submodules
        that have been imported.
        """
        if self.modname in _registered_mods:
            return
        package_prefix = self.modname + '.'
        for key_modname, elements in _preregistered_elements.items():
            # check for prefix to allow definition in submodules of a package
            if key_modname in _registered_mods:
                continue
            if key_modname == self.modname or key_modname.startswith(package_prefix):
                logger.debug('Registering schema and controls for {0}'.format(key_modname))
                for obj in elements:
                    obj.register()
                _registered_mods.add(self.modname)



#
# This file is part of pyasn1 software.
#
# Copyright (c) 2005-2019, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/pyasn1/license.html
#


class PyAsn1Error(Exception):
    """Create pyasn1 exception object

    The `PyAsn1Error` exception represents generic, usually fatal, error.
    """


class PyAsn1StringError(PyAsn1Error, UnicodeError):
    """Create pyasn1 exception object

    The `PyAsn1StringError` exception is a base class for errors relating to
    string encoding/decoding and other related problems
    """
    def __init__(self, message, unicode_error=None):
        if isinstance(unicode_error, UnicodeError):
            UnicodeError.__init__(self, *unicode_error.args)
        PyAsn1Error.__init__(self, message)


class PyAsn1StringDecodeError(PyAsn1StringError, UnicodeDecodeError):
    """Create pyasn1 exception object

    The `PyAsn1StringDecodeError` exception represents a failure to decode
    underlying bytes values to a string
    """


class PyAsn1StringEncodeError(PyAsn1StringError, UnicodeEncodeError):
    """Create pyasn1 exception object

    The `PyAsn1StringEncodeError` exception represents a failure to encode
    underlying string value as bytes
    """


class ValueConstraintError(PyAsn1Error):
    """Create pyasn1 exception object

    The `ValueConstraintError` exception indicates an ASN.1 value
    constraint violation.
    """


class SubstrateUnderrunError(PyAsn1Error):
    """Create pyasn1 exception object

    The `SubstrateUnderrunError` exception indicates insufficient serialised
    data on input of a de-serialization routine.
    """



"""Translations of ABNF specs to regex from RFC 4512

https://tools.ietf.org/html/rfc4512
"""

from __future__ import absolute_import

ALPHA = r'[A-Za-z]'
DIGIT = r'[0-9]'
HEX = r'[0-9A-Fa-f]'

SP = r' +'
WSP = r' *'

keychar = r'[A-Za-z0-9-]' # ALPHA / DIGIT / HYPHEN
keystring = ALPHA + keychar + r'*'

numericoid = DIGIT + r'+(?:\.' + DIGIT + r'+)+'
descr = keystring
oid = r'(?:' + descr + r'|' + numericoid + r')'

qdescr = r"'" + descr + r"'"
qdescrlist = qdescr + r'(?:' + SP + qdescr + r')*'
qdescrs = r'(?:' + qdescr + r'|\(' + WSP + qdescrlist + WSP + r'\))'

QQ = r'\\27'
QS = r'\\5[Cc]'
QUTF8 = r"[^'\\]"

dstring = r'(?:' + QS + r'|' + QQ + r'|' + QUTF8 + r')+'
qdstring = r"'" + dstring + r"'"
qdstringlist = qdstring + r'(?:' + SP + qdstring + r')*'
qdstrings = r'(?:' + qdstring + r'|\(' + WSP + qdstringlist + WSP + r'\))'

oidlist = oid + r'(?:' + WSP + r'\$' + WSP + oid + r')*'
oids = r'(?:' + oid + r'|\(' + WSP + oidlist + WSP + r'\))'

noidlen = numericoid + r'(\{[0-9]+\})?'

xstring = r'(?:X-[A-Za-z_-]+)'
extensions = r'(?P<extensions> ' + xstring + r' ' + qdstrings + r')?'

ObjectClassDescription = (
    r'\(' + WSP +
    r'(?P<oid>' + numericoid + r')' +
    r'(?:' + SP + r'NAME' + SP + r'(?P<name>' + qdescrs + r'))?' +
    r'(?:' + SP + r'DESC' + SP + r'(?P<desc>' + qdstring + r'))?' +
    r'(?P<obsolete>' + SP + r'OBSOLETE)?' +
    r'(?:' + SP + r'SUP' + SP + r'(?P<superclass>' + oids + r'))?' +
    r'(?:' + SP + r'(?P<kind>ABSTRACT|STRUCTURAL|AUXILIARY))?' +
    r'(?:' + SP + r'MUST' + SP + r'(?P<must>' + oids + r'))?' +
    r'(?:' + SP + r'MAY' + SP + r'(?P<may>' + oids + r'))?' +
    extensions + WSP + r'\)'
)

AttributeTypeDescription = (
    r'\(' + WSP +
    r'(?P<oid>' + numericoid + r')' +
    r'(?:' + SP + r'NAME' + SP + r'(?P<name>' + qdescrs + r'))?' +
    r'(?:' + SP + r'DESC' + SP + r'(?P<desc>' + qdstring + r'))?' +
    r'(?P<obsolete>' + SP + r'OBSOLETE)?' +
    r'(?:' + SP + r'SUP' + SP + r'(?P<supertype>' + oid + r'))?' +
    r'(?:' + SP + r'EQUALITY' + SP + r'(?P<equality>' + oid + r'))?' +
    r'(?:' + SP + r'ORDERING' + SP + r'(?P<ordering>' + oid + r'))?' +
    r'(?:' + SP + r'SUBSTR' + SP + r'(?P<substr>' + oid + r'))?' +
    r'(?:' + SP + r'SYNTAX' + SP + r'(?P<syntax>' + noidlen + r'))?' +
    r'(?P<single_value>' + SP + r'SINGLE-VALUE)?' +
    r'(?P<collective>' + SP + r'COLLECTIVE)?' +
    r'(?P<no_user_mod>' + SP + r'NO-USER-MODIFICATION)?' +
    r'(?:' + SP + r'USAGE' + SP + r'(?P<usage>userApplications|directoryOperation|distributedOperation|dSAOperation))?' +
    extensions + WSP + r'\)'
)

MatchingRuleDescription = (
    r'\(' + WSP +
    r'(?P<oid>' + numericoid + r')' +
    r'(?:' + SP + r'NAME' + SP + qdescrs + r')?' +
    r'(?:' + SP + r'DESC' + SP + qdstring + r')?' +
    r'(?:' + SP + r'OBSOLETE)?' +
    r'(?:' + SP + r'SYNTAX' + SP + oids + r')' +
    extensions + WSP + r'\)'
)

MatchingRuleUseDescription = (
    r'\(' + WSP +
    r'(?P<oid>' + numericoid + r')' +
    r'(?:' + SP + r'NAME' + SP + qdescrs + r')?' +
    r'(?:' + SP + r'DESC' + SP + qdstring + r')?' +
    r'(?:' + SP + r'OBSOLETE)?' +
    r'(?:' + SP + r'APPLIES' + SP + oids + r')' +
    extensions + WSP + r'\)'
)

SyntaxDescription = (
    r'\(' + WSP +
    r'(?P<oid>' + numericoid + r')' +
    r'(?:' + SP + r'DESC' + SP + qdstring + r')?' +
    extensions + WSP + r'\)'
)

DITContentRuleDescription = (
    r'\(' + WSP +
    r'(?P<oid>' + numericoid + r')' +
    r'(?:' + SP + r'NAME' + SP + qdescrs + r')?' +
    r'(?:' + SP + r'DESC' + SP + qdstring + r')?' +
    r'(?:' + SP + r'OBSOLETE' + SP + r')?' +
    r'(?:' + SP + r'AUX' + SP + oids + r')?' +
    r'(?:' + SP + r'MUST' + SP + oids + r')?' +
    r'(?:' + SP + r'NOT' + SP + oids + r')?' +
    extensions + WSP + r'\)'
)

ruleid = r'[0-9]+'
ruleidlist = ruleid + r'(?:' + SP + ruleid + r')*'
ruleids = r'(?:' + ruleid + r'|\(' + WSP + ruleidlist + WSP + r'\))'

DITStructureRuleDescription = (
    r'\(' + WSP +
    r'(?:' + ruleid +  r')' +
    r'(?:' + SP + r'NAME' + SP + qdescrs + r')?' +
    r'(?:' + SP + r'DESC' + SP + qdstring + r')?' +
    r'(?:' + SP + r'OBSOLETE' + SP + r')?' +
    r'(?:' + SP + r'FORM' + SP + oid + r')' +
    r'(?:' + SP + r'SUP' + SP + ruleids + r')?' +
    extensions + WSP + r'\)'
)

NameFormDescription = (
    r'\(' +  WSP +
    r'(?:' + numericoid + r')' +
    r'(?:' + SP + r'NAME' + SP + qdescrs + r')?' +
    r'(?:' + SP + r'DESC' + SP + qdstring + r')?' +
    r'(?:' + SP + r'OBSOLETE)?' +
    r'(?:' + SP + r'OC' + SP + oid + r')' +
    r'(?:' + SP + r'MUST' + SP + oids + r')' +
    r'(?:' + SP + r'MAY' + SP + oids + r')?' +
    extensions + WSP + r'\)'
)



from __future__ import absolute_import
from . import rfc4512
from . import rules
from . import utils
from .exceptions import LDAPSchemaError, InvalidSyntaxError
from .protoutils import parse_qdescrs
from .utils import CaseIgnoreDict

import logging
import re

logger = logging.getLogger(__name__)

_re_attr_type = re.compile(utils.re_anchor(rfc4512.AttributeTypeDescription))

_oid_attribute_types = {}
_name_attribute_types = CaseIgnoreDict()


def get_attribute_type(ident):
    """Get an instance of :class:`AttributeType` associated with either a name or OID.

    :param str ident: Either the numeric OID of the desired attribute type spec or any one of its specified names
    :return: The AttributeType containing a parsed specification
    :rtype: AttributeType
    """
    if ident[0].isdigit():
        return _oid_attribute_types[ident]
    else:
        try:
            return _name_attribute_types[ident]
        except KeyError:
            return DefaultAttributeType(ident)


class AttributeType(object):
    """Parses an LDAP attribute type specification and implements supertype inheritance.

    Each instantiation registers the names and OIDs specified so that the spec can be accessed using
    :func:`get_attribute_type`.

    See the :mod:`laurelin.ldap.schema` module source for example usages.

    :param str spec: The LDAP specification for an Attribute Type.
    :raises: LDAPSchemaError:
     * if the specification is invalid
     * if the OID has already been defined
     * if one of the names has already been defined

    :var str oid: The OID of the attribute type
    :var tuple(str) names: A tuple containing all possible names for the attribute type
    :var str supertype: The specified supertype. If the spec does not define optional properties, they will pass through
                        into the supertype.
    :var str equality_oid: The OID of the equality matching rule
    :var str syntax_oid: The OID of the syntax matching rule
    :var int syntax_length: The suggested maximum length of a value
    :var bool obsolete: The type has been flagged as obsolete. Will cause a warning from the :class:`SchemaValidator` if
                        an obsolete attribute type is used.
    :var bool single_value: The attribute may only have one value.
    :var bool collective: The attribute has been marked collective.
    :var bool no_user_mod: The attribute may not be modified by users (e.g., for operational attributes). Will cause a
                           validation failure from the :class:`SchemaValidator` if a write operation is attempted on
                           attribute types with this property set to True.
    :var str usage: A string describing the attribute's usage. May be one of `userApplications`, `directoryOperation`,
                    `distributedOperation`, or `dSAOperation`.
    """
    def __init__(self, spec):
        spec = utils.collapse_whitespace(spec).strip()
        m = _re_attr_type.match(spec)
        if not m:
            raise LDAPSchemaError('Invalid attribute type specification')

        self.oid = m.group('oid')
        if not self.oid:
            raise LDAPSchemaError('No OID defined for attribute type')

        self.names = parse_qdescrs(m.group('name'))
        if not self.names:
            raise LDAPSchemaError('No names defined for attribute type {0}'.format(self.oid))

        self.supertype = m.group('supertype')
        if self.supertype:
            self.supertype = self.supertype.strip()

        equality = m.group('equality')
        if equality is not None:
            self.equality_oid = equality
        elif not self.supertype:
            self.equality_oid = None

        # Note: ordering and substring matching not currently implemented
        # specs stored in m.group('ordering') and m.group('substr')

        syntax = m.group('syntax')
        if syntax is not None:
            syntax_noidlen = syntax.split('{')
            self.syntax_oid = syntax_noidlen[0]
            if len(syntax_noidlen) > 1:
                self.syntax_length = int(syntax_noidlen[1].strip('}'))
            else:
                self.syntax_length = -1
        elif not self.supertype:
            self.syntax_oid = None
            self.syntax_length = -1

        obsolete = m.group('obsolete')
        if obsolete is not None:
            self.obsolete = bool(obsolete)
        elif not self.supertype:
            self.obsolete = False

        single_value = m.group('single_value')
        if single_value is not None:
            self.single_value = bool(single_value)
        elif not self.supertype:
            self.single_value = False

        collective = m.group('collective')
        if collective is not None:
            self.collective = bool(collective)
        elif not self.supertype:
            self.collective = False

        no_user_mod = m.group('no_user_mod')
        if no_user_mod is not None:
            self.no_user_mod = bool(no_user_mod)
        elif not self.supertype:
            self.no_user_mod = False

        usage = m.group('usage')
        if usage:
            self.usage = usage
        elif not self.supertype:
            self.usage = 'userApplications'

    def register(self):
        # register OID
        if self.oid in _oid_attribute_types:
            raise LDAPSchemaError('Duplicate attribute type OID {0}'.format(self.oid))
        _oid_attribute_types[self.oid] = self

        # register name(s)
        for name in self.names:
            if name in _name_attribute_types:
                raise LDAPSchemaError('Duplicate attribute type name {0}'.format(name))
            _name_attribute_types[name] = self

    @property
    def syntax(self):
        """Gets the :class:`SyntaxRule` for this attribute type."""
        if not self.syntax_oid:
            raise LDAPSchemaError('Attribute type {0} does not have a defined syntax'.format(self.oid))
        return rules.get_syntax_rule(self.syntax_oid)

    @property
    def equality(self):
        """Gets the :class:`EqualityMatchingRule` for this attribute type."""
        if not self.equality_oid:
            raise LDAPSchemaError('Attribute type {0} does not have a defined equality matching rule'.format(self.oid))
        return rules.get_matching_rule(self.equality_oid)

    def __getattr__(self, name):
        if self.supertype:
            return getattr(get_attribute_type(self.supertype), name)
        else:
            raise AttributeError("No attribute named '{0}' and no supertype specified for attr {1}".format(
                                 name, self.oid))

    def validate(self, value):
        """Validate a value according to the attribute type's syntax rule.

        :param str value: The potential attribute value
        :return: A truthy value.
        :raises InvalidSyntaxError: if the value is invalid.
        """
        if self.syntax_length > -1:
            length = len(value)
            if length > self.syntax_length:
                raise InvalidSyntaxError('Length {0} greater than allowed {1}'.format(length, self.syntax_length))
        return self.syntax.validate(value)

    def index(self, value_list, assertion_value):
        """Finds the index of a value in a list of attribute values. Raises a
         ValueError if the value is not found in the list. Assumes values in
         value_list are already validated.

         :param list[str] value_list: The list of attribute values. Assumes values are already validated.
         :param str assertion_value: The value to look for in ``value_list``.
         :return: The index of ``assertion_value`` in ``value_list``.
         :rtype: int
         :raises ValueError: if ``assertion_value`` is not found or if ``value_list`` is empty.
         :raises InvalidSyntaxError: if ``assertion_value`` does not meet the syntax requirements of this attribute type
        """
        if not value_list:
            raise ValueError('empty value_list')
        self.validate(assertion_value)
        assertion_value = self.equality.prepare(assertion_value)
        for i, val in enumerate(value_list):
            val = self.equality.prepare(val)
            if self.equality.do_match(val, assertion_value):
                return i
        raise ValueError('assertion_value not found')

    def __repr__(self):
        return '<{0} "{1}">'.format(self.__class__.__name__, self.names[0])


## Defaults used when an attribute type is undefined

class DefaultSyntaxRule(object):
    """The default syntax rule to use for undefined attribute types.

    Users should probably never instantiate this.
    """
    def validate(self, s):
        """Allow all values"""
        pass


class DefaultMatchingRule(object):
    """The default matching rule to use for undefined attribute types.

    Users should probably never instantiate this.
    """
    def validate(self, value):
        """Allow all values"""
        return True

    def prepare(self, a):
        """Do nothing to prepare"""
        return a

    def do_match(self, a, b):
        """Require strict equality"""
        return (a == b)

    def match(self, a, b):
        """Do the match"""
        return self.do_match(a, b)


class DefaultAttributeType(AttributeType):
    """The default attribute type returned by :func:`get_attribute_type` when the requested attribute type is
    undefined.

    Essentially behaves as an unrestricted case-sensitive attribute type.

    Users should probably never instantiate this.
    """
    def __init__(self, name=None):
        logger.debug('Using DefaultAttributeType for name={0}'.format(name))
        self.oid = None
        self.names = (name,)
        self._equality = DefaultMatchingRule()
        self._syntax = DefaultSyntaxRule()
        self.obsolete = False
        self.single_value = False
        self.collective = False
        self.no_user_mod = False
        self.usage = 'userApplications'
        self.supertype = None

    @property
    def syntax(self):
        return self._syntax

    @property
    def equality(self):
        return self._equality

    def index(self, value_list, assertion_value):
        return list.index(value_list, assertion_value)



# Automatically generated by scripts/generate_extension_properties.py
# **DO NOT MODIFY - CHANGES WILL BE OVERWRITTEN**

from .base import ExtensibleClass


class LDAPExtensions(ExtensibleClass):

    @property
    def netgroups(self):
        """The built-in NIS netgroups extension

        :rtype: laurelin.extensions.netgroups.LaurelinLDAPExtension
        """
        return self._get_extension_instance('netgroups')



"""Contains utilities for performing object modification"""

from __future__ import absolute_import
from .attributetype import get_attribute_type
from .rfc4511 import Operation
from .constants import DELETE_ALL
import six


class Mod(object):
    """Describes a single modify operation"""
    ADD = Operation('add')
    REPLACE = Operation('replace')
    DELETE = Operation('delete')

    @staticmethod
    def op_to_string(op):
        """Convert one of the :class:`Mod` constants to a string, e.g. "ADD", "REPLACE", "DELETE"."""
        if op == Mod.ADD:
            return 'ADD'
        elif op == Mod.REPLACE:
            return 'REPLACE'
        elif op == Mod.DELETE:
            return 'DELETE'
        else:
            raise ValueError()

    @staticmethod
    def string(op):
        """Translte LDIF changetype strings to constant. e.g. "replace" -> :attr:`.Mod.REPLACE`"""
        if op == 'add':
            return Mod.ADD
        elif op == 'replace':
            return Mod.REPLACE
        elif op == 'delete':
            return Mod.DELETE
        else:
            raise ValueError()

    def __init__(self, op, attr, vals):
        if (op != Mod.ADD) and (op != Mod.REPLACE) and (op != Mod.DELETE):
            raise ValueError()
        if vals is not DELETE_ALL and not isinstance(vals, list):
            vals = [vals]
        if (op == Mod.ADD) and (len(vals) == 0):
            raise ValueError('No values to add')
        self.op = op
        self.attr = attr
        self.vals = vals

    def __repr__(self):
        if self.vals:
            vals = str(self.vals)
        else:
            vals = 'DELETE_ALL'
        return 'Mod({0}, {1}, {2})'.format(repr(self.op), repr(self.attr), vals)


def _mod_op_repr(operation_obj):
    """Uses laurelin constant name representation for mod operations"""
    try:
        intval = int(operation_obj)
        name = operation_obj.namedValues.getName(intval)
        if name == 'add':
            return 'Mod.ADD'
        elif name == 'replace':
            return 'Mod.REPLACE'
        elif name == 'delete':
            return 'Mod.DELETE'
        else:
            return '{0}({1})'.format(operation_obj.__class__.__name__, repr(name))
    except Exception:
        return '{0}(<schema object>)'.format(operation_obj.__class__.__name__)


Operation.__repr__ = _mod_op_repr

def Modlist(op, attrs_dict):
    """Generate a modlist from a dictionary"""

    if not isinstance(attrs_dict, dict):
        raise TypeError('attrs_dict must be dict')
    modlist = []
    for attr, vals in six.iteritems(attrs_dict):
        modlist.append(Mod(op, attr, vals))
    return modlist


## Smart modlist functions which will prevent errors


def AddModlist(cur_attrs, new_attrs):
    """Generate a modlist to add only new attribute values that are not known to exist"""

    if not isinstance(cur_attrs, dict):
        raise TypeError('cur_attrs must be dict')
    if not isinstance(new_attrs, dict):
        raise TypeError('new_attrs must be dict')
    add_attrs = {}
    for attr, vals in six.iteritems(new_attrs):
        if attr in cur_attrs:
            attr_type = get_attribute_type(attr)
            for val in vals:
                try:
                    attr_type.index(cur_attrs[attr], val)
                    # attribute value already exists, do nothing
                except ValueError:
                    # attribute value does not exist, add it
                    if attr not in add_attrs:
                        add_attrs[attr] = []
                    add_attrs[attr].append(val)
        else:
            add_attrs[attr] = vals
    return Modlist(Mod.ADD, add_attrs)


def DeleteModlist(cur_attrs, del_attrs):
    """Generate a modlist to delete only attribute values that are known to exist"""

    if not isinstance(del_attrs, dict):
        raise TypeError('cur_attrs must be dict')
    if not isinstance(del_attrs, dict):
        raise TypeError('del_attrs must be dict')
    _del_attrs = {}
    for attr, vals in six.iteritems(del_attrs):
        if attr in cur_attrs:
            if not vals:
                _del_attrs[attr] = vals
            else:
                attr_type = get_attribute_type(attr)
                for val in vals:
                    try:
                        attr_type.index(cur_attrs[attr], val)
                        # attribute value exists, delete it
                        if attr not in _del_attrs:
                            _del_attrs[attr] = []
                        _del_attrs[attr].append(val)
                    except ValueError:
                        # attribute value does not exist, do nothing
                        pass
    return Modlist(Mod.DELETE, _del_attrs)



from __future__ import absolute_import
from .attributetype import get_attribute_type
from .extensible import extensions


class AttrValueList(list):
    """List that follows schema matching rules for the ``in`` operator and other related methods.

    :param str attr: The attribute name or type identifier
    :param list[str] values: Initial values for the list
    """

    def __init__(self, attr, values):
        self.attr = attr
        list.__init__(self, values)

    def __contains__(self, value):
        try:
            self.index(value)
            return True
        except ValueError:
            return False

    def index(self, value, *args, **kwds):
        """Find the index of value or raise a ValueError if not found.
        The stock start/end arguments are ignored since a list of attribute
        values is defined to have exactly zero or one unique matching values.

        :param str value: The value to find
        :return: The index of the value
        :rtype: int
        :raises ValueError: if the value is not found or if the list has no values
        """
        extensions.base_schema.require()
        attr_type = get_attribute_type(self.attr)
        return attr_type.index(self, value)

    def count(self, value):
        """Count the number of occurrences of ``value``. Since attribute value lists are defined to only have at most
        one unique copy of any value, this will always return 0 or 1.

        :param str value: The value to count
        :return: The number of occurrences of ``value``, 1 or 0.
        :rtype: int
        :raises ValueError: if the value is not found or if the list has no values
        """
        try:
            self.index(value)
            return 1
        except ValueError:
            return 0

    def remove(self, value):
        """Remove ``value`` from the list if present.

        :param str value: The value to remove
        :rtype: None
        :raises ValueError: if the value is not found or if the list has no values
        """
        i = self.index(value)
        del self[i]



import six


class Validator(object):
    """Abstract base class for a validator. All validators must inherit from here and ensure the public interface is
    fully implemented.
    """
    def __init__(self):
        self.ldap_conn = None

    def validate_object(self, obj, write=True):
        """Validate an object when all attributes are present.

        By default, validate all attributes on the object.

        :param LDAPObject obj: An LDAP object with all attributes defined
        :param bool write: True if we are validating a write operation to the database
        :return: None
        :raises LDAPValidationError: if the object is invalid in any way
        """
        for attr, values in six.iteritems(obj):
            self._validate_attribute(attr, values, write)

    def validate_modify(self, dn, modlist, current):
        """Validate a modify operation.

        By default, validate all attributes for writing.

        :param str dn: The DN of the object being modified
        :param list[Mod] modlist: The list of modify operations to be performed this transaction
        :param current: The known state of the object prior to modification
        :type current: LDAPObject or None
        :return: None
        :raises LDAPValidationError: if any modify operation is invalid
        """
        for mod in modlist:
            if mod.vals:
                self._validate_attribute(mod.attr, mod.vals, True)

    def _validate_attribute(self, attr_name, values, write):
        """Validate a single attribute and list of values.

        :param str attr_name: The name of the attribute, or more formally the attribute type, being modified
        :param list[str] values: The new list of values or values to be deleted
        :param bool write: True if we are validating a write operation
        :return: None
        :raises LDAPValidationError: if the attribute is invalid in any way
        """
        raise NotImplementedError()


class DisabledValidationContext(object):
    """This should be created by calling :meth:`.LDAP.disable_validation` and never directly instantiated."""

    def __init__(self, ldap, disabled_validators=None):
        self.ldap = ldap
        self.orig_validators = []
        self.disabled_validators = set()
        if disabled_validators:
            for val_spec in disabled_validators:
                if isinstance(val_spec, six.string_types):
                    self.disabled_validators.add(val_spec)
                elif issubclass(val_spec, Validator):
                    self.disabled_validators.add(val_spec.__name__)
                else:
                    raise TypeError('Invalid disabled validator spec, must be string class name or Validator subclass')

    def __enter__(self):
        new_validators = []
        for v in self.ldap.validators:
            self.orig_validators.append(v)
            if not self.disabled_validators:
                continue
            if v.__class__.__name__ in self.disabled_validators:
                continue
            new_validators.append(v)
        self.ldap.validators = new_validators

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.ldap.validators = self.orig_validators



class LDAPError(Exception):
    """Base class for all exceptions raised by laurelin"""
    pass


class Abandon(Exception):
    """Can be raised to cleanly exit a context manager and abandon unread results"""
    pass


class LDAPWarning(Warning):
    """Generic LDAP warning category"""
    pass


class LDAPUnicodeWarning(LDAPWarning, UnicodeWarning):
    """Warning category for unicode issues relating to LDAP"""
    pass


class LDAPUnsolicitedMessage(Exception):
    """Raised when a message with ID 0 is returned from the server

    This may indicate an incompatability between laurelin and your server distribution and thus is outside the normal
    exception inheritance chain.
    """
    def __init__(self, lm, exc_msg):
        self.ldap_message = lm
        Exception.__init__(self, exc_msg)


class LDAPExtensionError(LDAPError):
    """Error occurred setting up an extension module"""
    pass


class LDAPSupportError(LDAPError):
    """A feature is not supported by the server"""
    pass


class LDAPSASLError(LDAPError):
    """Error occurred involving the SASL client"""
    pass


class LDAPConnectionError(LDAPError):
    """Error occurred creating connection to the LDAP server"""
    pass


class TagError(LDAPError):
    """Error with an object tag"""
    pass


class UnexpectedResponseType(LDAPError):
    """The response did not contain the expected protocol operation"""
    pass


class UnexpectedSearchResults(LDAPError):
    """Base class for unhandled search result situations"""
    pass


class NoSearchResults(UnexpectedSearchResults):
    """Got no search results when one or more was required"""
    pass


class MultipleSearchResults(UnexpectedSearchResults):
    """Got multiple search results when exactly one was required"""
    pass


class InvalidBindState(LDAPError):
    """Base class for exceptions related to bind state"""
    pass


class ConnectionAlreadyBound(InvalidBindState):
    """Only raised by LDAP.*Bind methods if the connection is already bound when called"""
    def __init__(self):
        LDAPError.__init__(self, 'The connection has already been bound')


class ConnectionUnbound(InvalidBindState):
    """Raised when any server operation is attempted after a connection is unbound/closed"""
    def __init__(self):
        LDAPError.__init__(self, 'The connection has been unbound')


class ProhibitedCharacterError(LDAPError):
    """Raised when a prohibited character is detected in RFC4518 string prep"""
    pass


class LDAPSchemaError(LDAPError):
    """Error relating to setting up the LDAP schema"""
    pass


class LDAPValidationError(LDAPError):
    """Raised when validation fails"""
    pass


class InvalidSyntaxError(LDAPValidationError):
    """Raised when syntax validation fails"""
    pass


class LDAPTransactionError(LDAPError):
    """Raised by actions not included in a modify transaction"""
    pass



# Automatically generated by scripts/generate_extension_properties.py
# **DO NOT MODIFY - CHANGES WILL BE OVERWRITTEN**

from .base import ExtensionsBase


class Extensions(ExtensionsBase):

    @property
    def base_schema(self):
        """The standard base schema from various RFCs

        :rtype: laurelin.extensions.base_schema.LaurelinExtension
        """
        return self._get_extension_instance('base_schema')

    @property
    def descattrs(self):
        """The built-in description attributes extension

        :rtype: laurelin.extensions.descattrs.LaurelinExtension
        """
        return self._get_extension_instance('descattrs')

    @property
    def netgroups(self):
        """The built-in NIS netgroups extension

        :rtype: laurelin.extensions.netgroups.LaurelinExtension
        """
        return self._get_extension_instance('netgroups')

    @property
    def paged_results(self):
        """Built-in extension defining standard paged results control for search

        :rtype: laurelin.extensions.pagedresults.LaurelinExtension
        """
        return self._get_extension_instance('paged_results')



"""Global constant classes."""

from __future__ import absolute_import
from .rfc4511 import (
    Scope as _Scope,
    DerefAliases as _DerefAliases,
)

_const_filter_syntax_standard = '__rfc4515_standard_filter_syntax__'
_const_filter_syntax_simple = '__laurelin_simplified_filter_syntax__'
_const_filter_syntax_unified = '__laurelin_unified_filter_syntax__'


class _FilterSyntaxSelection(object):
    def __init__(self, const):
        self.const = const

    def __repr__(self):
        if self.const == _const_filter_syntax_simple:
            return 'FilterSyntax.SIMPLE'
        elif self.const == _const_filter_syntax_standard:
            return 'FilterSyntax.STANDARD'
        elif self.const == _const_filter_syntax_unified:
            return 'FilterSyntax.UNIFIED'
        else:
            return repr(self.const)


class FilterSyntax:
    """Filter syntax selection constants. Used to determine which filter syntax to use when parsing a search filter."""
    STANDARD = _FilterSyntaxSelection(_const_filter_syntax_standard)
    SIMPLE = _FilterSyntaxSelection(_const_filter_syntax_simple)
    UNIFIED = _FilterSyntaxSelection(_const_filter_syntax_unified)

    @staticmethod
    def string(str):
        """Convert filter syntax string to constant"""
        str = str.upper()
        return getattr(FilterSyntax, str)


class _DeleteAllAttrs(object):
    """Sentinel object used to delete all attributes in replace or delete"""
    def __bool__(self):
        return False

    def __nonzero__(self):
        return False

    def __len__(self):
        return 0

    def __repr__(self):
        return '<delete all values>'


DELETE_ALL = _DeleteAllAttrs()


class Scope:
    """Scope constants. These instruct the server how far to take a search, relative to the base object"""

    BASE = _Scope('baseObject')
    """Only search the base object"""

    ONELEVEL = _Scope('singleLevel')
    ONE = ONELEVEL
    """Search the base object and its immediate children"""

    SUBTREE = _Scope('wholeSubtree')
    SUB = SUBTREE
    """Search the base object and all of its dscendants"""

    @staticmethod
    def string(str):
        """translate RFC4516 URL scope strings to constant"""
        str = str.lower()
        if str == 'base':
            return Scope.BASE
        elif str == 'one':
            return Scope.ONELEVEL
        elif str == 'sub':
            return Scope.SUBTREE
        else:
            raise ValueError()

    @staticmethod
    def constant(c):
        """translate constants to RFC4516 URL scope string"""
        if c == Scope.BASE:
            return 'base'
        elif c == Scope.ONE:
            return 'one'
        elif c == Scope.SUB:
            return 'sub'
        else:
            raise ValueError()


def _scope_repr(scope_obj):
    """Uses laurelin constant name representation for scope"""
    try:
        intval = int(scope_obj)
        name = scope_obj.namedValues.getName(intval)
        if name == 'wholeSubtree':
            return 'Scope.SUB'
        elif name == 'singleLevel':
            return 'Scope.ONE'
        elif name == 'baseObject':
            return 'Scope.BASE'
        else:
            return '{0}({1})'.format(scope_obj.__class__.__name__, repr(name))
    except Exception:
        return '{0}(<schema object>)'.format(scope_obj.__class__.__name__)


_Scope.__repr__ = _scope_repr


class DerefAliases:
    """DerefAliases constants. These instruct the server when to automatically resolve an alias object, rather than
       return the alias object itself
    """

    NEVER = _DerefAliases('neverDerefAliases')
    """always return the alias object"""

    SEARCH = _DerefAliases('derefInSearching')
    """dereferences search results, but not the base object itself"""

    BASE = _DerefAliases('derefFindingBaseObj')
    """dereferences the search base object, but not search results"""

    ALWAYS = _DerefAliases('derefAlways')
    """dereferences both the search base object and results"""


def _deref_repr(deref_obj):
    """Uses laurelin constant name representation for deref aliases"""
    try:
        intval = int(deref_obj)
        name = deref_obj.namedValues.getName(intval)
        if name == 'neverDerefAliases':
            return 'DerefAliases.NEVER'
        elif name == 'derefInSearching':
            return 'DerefAliases.SEARCH'
        elif name == 'derefFindingBaseObj':
            return 'DerefAliases.BASE'
        elif name == 'derefAlways':
            return 'DerefAliases.ALWAYS'
        else:
            return '{0}({1})'.format(deref_obj.__class__.__name__, repr(name))
    except Exception:
        return '{0}(<schema object>)'.format(deref_obj.__class__.__name__)


_DerefAliases.__repr__ = _deref_repr



from __future__ import absolute_import
from . import rfc4511
from .exceptions import UnexpectedResponseType
from .pyasn1.error import PyAsn1Error
import logging
import six
from six.moves import range

# Commonly reused protocol objects
V3 = rfc4511.Version(3)
EMPTY_DN = rfc4511.LDAPDN('')
RESULT_saslBindInProgress = rfc4511.ResultCode('saslBindInProgress')
RESULT_success = rfc4511.ResultCode('success')
RESULT_noSuchObject = rfc4511.ResultCode('noSuchObject')
RESULT_compareTrue = rfc4511.ResultCode('compareTrue')
RESULT_compareFalse = rfc4511.ResultCode('compareFalse')
RESULT_referral = rfc4511.ResultCode('referral')

logger = logging.getLogger(__name__)


def pack(mid, op, obj, controls=None):
    """Pack an object into an LDAPMessage envelope"""
    lm = rfc4511.LDAPMessage()
    lm.setComponentByName('messageID', rfc4511.MessageID(mid))
    po = rfc4511.ProtocolOp()
    po.setComponentByName(op, obj)
    lm.setComponentByName('protocolOp', po)
    if controls:
        lm.setComponentByName('controls', controls)
    return lm


def unpack(op, ldap_message):
    """Unpack an object from an LDAPMessage envelope"""
    mid = ldap_message.getComponentByName('messageID')
    po = ldap_message.getComponentByName('protocolOp')
    controls = ldap_message.getComponentByName('controls')
    got_op = po.getName()
    if got_op == op:
        ret = po.getComponent()
        if ret.isValue:
            return mid, ret, controls
    raise UnexpectedResponseType('Got {0} but expected {1}'.format(got_op, op))


def seq_to_list(seq):
    """Convert a pyasn1 sequence to a list of strings"""
    ret = []
    for i in range(len(seq)):
        try:
            ret.append(six.text_type(seq.getComponentByPosition(i)))
        except UnicodeDecodeError:
            ret.append(six.binary_type(seq.getComponentByPosition(i)))
    return ret


def parse_qdescrs(spec):
    """Parse an rfc4512.qdescrs to a tuple"""
    if spec is None:
        return ()
    return tuple(qdescr.strip("'") for qdescr in spec.strip('( )').split(' '))


def get_string_component(obj, name):
    """Try to get a string component from a PyASN1 object, or an empty string on error"""
    try:
        comp = obj.getComponentByName(name)
        return six.text_type(comp)
    except PyAsn1Error:
        logger.debug("Returning empty string for {0} due to PyAsn1Error".format(name))
        return ''



"""RFC 4514: String Representation of Distinguished Names

https://tools.ietf.org/html/rfc4514
"""

from __future__ import absolute_import
from . import rfc4512

## Section 3


hexpair = rfc4512.HEX + r'{2}'
hexstring = r'#(?:' + hexpair + r')+'

escaped = r'"+,;<>\0\\'
special = escaped + ' #='

pair = r'\\([' + special + r']|' + hexpair + r')'

stringchar = r'[^' + escaped + r']'
leadchar = r'[^' + special + r']'
trailchar = r'[^' + escaped + r' ]'

string = (
    r'(?:' + leadchar + r'|' + pair + r')' +
    r'(?:' + stringchar + r'|' + pair + r')*' +
    r'(?:' + trailchar + r'|' + pair + r')?'
)

attributeValue = r'(?:' + string + r'|' + hexstring + r')'
attributeType = r'(?:' + rfc4512.descr + r'|' + rfc4512.numericoid + r')'
attributeTypeAndValue = attributeType + r'=' + attributeValue

relativeDistinguishedName = attributeTypeAndValue + r'(?:\+' + attributeTypeAndValue + r')*'
distinguishedName = relativeDistinguishedName + r'(?:,' + relativeDistinguishedName + r')*'



"""Imports and defines the core of the public API"""

from __future__ import absolute_import

from .attributetype import get_attribute_type, AttributeType
from .base import LDAP, LDAPURI
from .constants import Scope, DerefAliases, DELETE_ALL, FilterSyntax
from .controls import Control, critical, optional
from .exceptions import LDAPError, NoSearchResults, Abandon
from .extensible import (
    extensions,
    add_extension,
    BaseLaurelinExtension,
    BaseLaurelinSchema,
    BaseLaurelinControls,
    BaseLaurelinLDAPExtension,
    BaseLaurelinLDAPObjectExtension,
    LaurelinTransiter,
    LaurelinRegistrar,
)
from .filter import escape as filter_escape
from .ldapobject import LDAPObject
from .modify import Mod
from .objectclass import get_object_class, ObjectClass, ExtensibleObjectClass
from .rules import SyntaxRule, RegexSyntaxRule, MatchingRule, EqualityMatchingRule
from .schema import SchemaValidator
from .validation import Validator
from .pyasn1.type import univ as _pyasn1_type_univ


def dc(domain):
    """Convert a DNS dotted domain name to a DN with domain components"""
    return ','.join(['dc={0}'.format(dc) for dc in domain.split('.')])


def domain(dc):
    """Convert a DN with domain components to a DNS dotted domain name"""
    return '.'.join([i.split('=')[1] for i in dc.split(',')])


__all__ = [
    'get_attribute_type',
    'AttributeType',
    'LDAP',
    'LDAPURI',
    'Scope',
    'DerefAliases',
    'DELETE_ALL',
    'FilterSyntax',
    'Control',
    'critical',
    'optional',
    'LDAPError',
    'NoSearchResults',
    'Abandon',
    'extensions',
    'add_extension',
    'BaseLaurelinExtension',
    'BaseLaurelinSchema',
    'BaseLaurelinControls',
    'BaseLaurelinLDAPExtension',
    'BaseLaurelinLDAPObjectExtension',
    'LaurelinTransiter',
    'LaurelinRegistrar',
    'filter_escape',
    'LDAPObject',
    'Mod',
    'get_object_class',
    'ObjectClass',
    'ExtensibleObjectClass',
    'SyntaxRule',
    'RegexSyntaxRule',
    'MatchingRule',
    'EqualityMatchingRule',
    'Validator',
    'SchemaValidator',
    'dc',
    'domain',
]


# make pyasn1 enumerated repr() consistant

def _enum_repr(enum_obj):
    try:
        intval = int(enum_obj)
        name = enum_obj.namedValues.getName(intval)
        namerepr = repr(name)
    except Exception:
        namerepr = '<schema object>'
    return '{0}({1})'.format(enum_obj.__class__.__name__, namerepr)


_pyasn1_type_univ.Enumerated.__repr__ = _enum_repr



"""Provides support for establishing an LDAP connection and environment via config files and dicts"""


from .base import LDAP
from .constants import Scope, FilterSyntax
from .validation import Validator
import json
import six
import yaml
from importlib import import_module


def _default_mapper(val):
    return val


def _validator_mapper(val):
    instances = []
    for v in val:
        if isinstance(v, six.string_types):
            modname, objname = v.rsplit('.', 1)
            mod = import_module(modname)
            vcls = getattr(mod, objname)
            vobj = vcls()
            instances.append(vobj)
        elif isinstance(v, Validator):
            instances.append(v)
        else:
            raise TypeError('"validators" list must be str or laurelin.ldap.validation.Validator')
    return instances


_connection_mappers = {
    'validators': _validator_mapper,
    'default_filter_syntax': FilterSyntax.string,
}

_global_mappers = {
    'DEFAULT_FILTER_SYNTAX': FilterSyntax.string,
}


def normalize_global_config_param(key):
    """Normalize a global config key. Does not check validity of the key.

    :param str key: User-supplied global config key
    :return: The normalized key formatted as an attribute of :class:`.LDAP`
    :rtype: str
    """
    key = key.upper()
    if not key.startswith('DEFAULT_'):
        key = 'DEFAULT_'+key
    return key


def set_global_config(global_config_dict):
    """Set the global defaults. The dict must be formatted as follows::

        {'global': {
            <config param>: <config value>,
         }
        }

    ``<config param>`` must match one of the ``DEFAULT_`` attributes on :class:`.LDAP`. The ``DEFAULT_`` prefix is
    optional and dict keys are case-insensitive. Any parameters not specified will keep the hard-coded default.

    :param dict global_config_dict: See above.
    :rtype: None
    :raises KeyError: if the dict is incorrectly formatted or contains unknown config parameters
    """
    bad = []
    for key, val in six.iteritems(global_config_dict['global']):
        orig_key = key
        key = normalize_global_config_param(key)
        if hasattr(LDAP, key):
            val = _global_mappers.get(key, _default_mapper)(val)
            setattr(LDAP, key, val)
        else:
            bad.append(orig_key)
    if bad:
        raise KeyError('Unknown global config keys: {0}'.format(', '.join(bad)))


def activate_extensions(config_dict):
    """Activate the specified extensions. The dict must be formatted as follows::

        {'extensions': [
            <module name>,
         ]
        }

    :param dict config_dict: See above.
    :rtype: None
    """
    for mod_name in config_dict['extensions']:
        LDAP.activate_extension(mod_name)


def _create_single_object(ldap, obj_config_dict):
    if 'rdn' in obj_config_dict and 'dn' in obj_config_dict:
        raise TypeError('Choose only one of "rdn" or "dn"')
    if 'rdn' in obj_config_dict:
        objmeth = ldap.base.obj
    elif 'dn' in obj_config_dict:
        objmeth = ldap.obj
    else:
        raise TypeError('missing required object key "rdn" or "dn"')
    key = 'relative_search_scope'
    if key in obj_config_dict:
        obj_config_dict[key] = Scope.string(obj_config_dict[key])
    key = 'tag'
    if key not in obj_config_dict:
        raise TypeError('missing required object key "{0}"'.format(key))
    objmeth(**obj_config_dict)


def create_connection(config_dict):
    """Create a new connection from a config dict formatted as follows::

        {'connection': {
            'start_tls': <bool>,  # optional, default False
            'simple_bind': {  # optional, default no bind; mutually exclusive with sasl_bind
                'username': <string username or bind dn>,
                'password': <string password>
            },
            'sasl_bind': {  # optional, default no bind, mutually exclusive with simple_bind
                'mech': <standard mech name>,
                <mech prop>: <mech value>,  # required props varies by mech
            },
            <constructor param>: <constructor value>,
         },
         'objects': [  # optional
            {'dn': <object dn>,  # OR...
             'rdn': <dn relative to connection base object>,
             'tag': <unique tag name>,
             <object param>: <object value>,
            },
            # ...
         ]
        }

    ``<constructor param>`` must be one of the :class:`.LDAP` constructor keyword arguments.

    For ``validators`` you can optionally give the full path to the validator to use as a string, e.g.
    ``['laurelin.ldap.schema.SchemaValidator']``

    For ``default_filter_syntax`` give one of the strings "STANDARD" or "SIMPLE" (case-insensitive).

    For objects (optional):

    * If the ``dn`` parameter is specified, it is taken as an absolute DN.
    * You can specify the ``rdn`` parameter instead to create the object as a child of the connection's base object (the
      base of the tree).
    * The ``tag`` parameter is required; this is how created objects are accessed (:meth:`.LDAP.tag`).
    * Additional ``<object param>`` will be passed as keywords to :meth:`.LDAP.obj`.
    * If ``relative_search_scope`` is specified, use one of the strings `base`, `one`, or `sub`.
    * The server will not be queried to create these objects, so they will have no local attributes. Call
      :meth:`.LDAPObject.refresh` if you need to query attributes.

    Note on binding: You can always manually call :meth:`.LDAP.simple_bind` or :meth:`.LDAP.sasl_bind` on the
    :class:`.LDAP` instance returned from this method if statically configuring bind credentials is not desirable.

    :param config_dict: See above.
    :return: The new LDAP instance with any objects created and tagged.
    :raises TypeError: if any required object parameters are missing
    """
    conn_config_dict = config_dict['connection']
    start_tls = conn_config_dict.pop('start_tls', False)
    simple_bind = conn_config_dict.pop('simple_bind', False)
    sasl_bind = conn_config_dict.pop('sasl_bind', False)
    if simple_bind and sasl_bind:
        raise TypeError('choose only one of simple_bind or sasl_bind')
    for key in _connection_mappers:
        if key in conn_config_dict:
            val = _connection_mappers[key](conn_config_dict[key])
            conn_config_dict[key] = val
    ldap = LDAP(**conn_config_dict)
    if start_tls:
        ldap.start_tls()
    if simple_bind:
        ldap.simple_bind(**simple_bind)
    if sasl_bind:
        ldap.sasl_bind(**sasl_bind)
    for obj_config_dict in config_dict.get('objects', []):
        _create_single_object(ldap, obj_config_dict)
    return ldap


def load_file(path, file_decoder=None):
    """Load a config file. Must decode to dict with all components described on other methods as optional sections/keys.
    A YAML example::

        extensions:
          - laurelin.extensions.descattrs
          - laurelin.extensions.netgroups
        global:
          SSL_CA_PATH: /etc/ldap/cacerts
          IGNORE_EMPTY_LIST: true
        connection:
          server: ldap://dir01.example.org
          start_tls: true
          simple_bind:
            username: testuser
            passowrd: testpassword
          connect_timeout: 30
        objects:
          - rdn: ou=people
            tag: posix_user_base
          - rdn: ou=groups
            tag: posix_group_base
          - rdn: ou=netgroups
            tag: netgroup_base

    :param path: A path to a config file. Provides support for YAML and JSON format, or you can specify your own decoder
                 that returns a dict.
    :param file_decoder: A callable returning a dict when passed a file-like object
    :return: The LDAP connection if one was defined, None otherwise
    :rtype: LDAP or None
    :raises RuntimeError: if an unsupported file extension was given without the ``file_decoder`` argument.
    """
    if file_decoder is None:
        if path.endswith('.yml') or path.endswith('.yaml'):
            file_decoder = yaml.load
        elif path.endswith('.json'):
            file_decoder = json.load
        else:
            raise RuntimeError('Unsupported file type, must be YAML or JSON, or specify file_decoder argument')
    with open(path) as f:
        config_dict = file_decoder(f)
    return load_config_dict(config_dict)


def load_config_dict(config_dict):
    """Load config parameters from a dictionary. Must be formatted in the same was as ``load_file``

    :param dict config_dict: The config dictionary. See format in ``load_file``.
    :return: The LDAP connection if one was defined, None otherwise
    :rtype: LDAP or None
    """
    if 'global' in config_dict:
        set_global_config(config_dict)
    if 'extensions' in config_dict:
        activate_extensions(config_dict)
    if 'connection' in config_dict:
        return create_connection(config_dict)



"""Support for structured description fields.

This implements the common pattern of storing arbitrary key=value data in description fields, but presents an
attribute-like interface to access and change them.

Example::

    from laurelin.ldap import LDAP
    LDAP.activate_extension('laurelin.extensions.descattrs')

    with LDAP() as ldap:
        result = ldap.base.get_child('cn=someObject')

        result.descattrs.add({'foo':['one', 'two']})
        print(result.format_ldif())
        # ...
        # description: foo=one
        # description: foo=two
        # ...

        attr_vals = result.descattrs.get_attr('foo')
        print(attr_vals)
        # ['one', 'two']

        result.descattrs.replace({'foo':['one','two','three']})
        result.descattrs.delete({'foo':['two']})

        attr_vals = result.descattrs.get_attr('foo')
        print(attr_vals)
        # ['one', 'three']

        print(result.format_ldif())
        # ...
        # description: foo=one
        # description: foo=three
        # ...

"""

from laurelin.ldap import BaseLaurelinLDAPObjectExtension, BaseLaurelinExtension
from laurelin.ldap.attrsdict import AttrsDict
import six

DESC_ATTR_DELIM = '='
"""Key/value delimiter in description attrs. If this character is not present in the description value, then it will be
considered unstructured and ignored.
"""


class LaurelinExtension(BaseLaurelinExtension):
    NAME = 'descattrs'


class LaurelinLDAPObjectExtension(BaseLaurelinLDAPObjectExtension):
    def __init__(self, parent):
        BaseLaurelinLDAPObjectExtension.__init__(self, parent)
        self.parent.refresh_missing(['description'])
        self._desc_dict = AttrsDict()
        self._unstructured_desc = set()
        for desc in self.parent.get_attr('description'):
            if DESC_ATTR_DELIM in desc:
                key, value = desc.split(DESC_ATTR_DELIM, 1)
                vals = self._desc_dict.setdefault(key, [])
                vals.append(value)
            else:
                self._unstructured_desc.add(desc)

    def __iter__(self):
        for attr in self._desc_dict:
            yield attr

    def __getitem__(self, attr):
        return self._desc_dict[attr]

    def __contains__(self, item):
        return item in self._desc_dict

    def _modify_desc_attrs(self, method, attrs_dict):
        """Perform modification to the object's description attributes.

        :param callable method: The method to call to modify the description attributes dictionary
        :param attrs_dict: Will be passed as the 2nd argument to ``method``.
        :type attrs_dict: dict(str, list[str]) or AttrsDict
        :rtype: None
        :raises RuntimeError: if there is no :class:`.LDAP` connection associated with this :class:`.LDAPObject`.
        """
        self.parent._require_ldap()
        method(self._desc_dict, attrs_dict)
        desc_strings = []
        for key, values in six.iteritems(self._desc_dict):
            for value in values:
                desc_strings.append(key + DESC_ATTR_DELIM + value)
        self.parent.replace_attrs({'description': desc_strings + list(self._unstructured_desc)})

    def add(self, attrs_dict):
        """Add new description attributes.

        :param attrs_dict: Dictionary of description attributes to add
        :type attrs_dict: dict(str, list[str]) or AttrsDict
        :rtype: None
        """
        self._modify_desc_attrs(_dict_mod_add, attrs_dict)

    def replace(self, attrs_dict):
        """Replace description attributes.

        :param attrs_dict: Dictionary of description attributes to set
        :type attrs_dict: dict(str, list[str]) or AttrsDict
        :rtype: None
        """
        self._modify_desc_attrs(_dict_mod_replace, attrs_dict)

    def delete(self, attrs_dict):
        """Delete description attributes.

        :param attrs_dict: Dictionary of description attributes to delete
        :type attrs_dict: dict(str, list[str]) or AttrsDict
        :rtype: None
        """
        self._modify_desc_attrs(_dict_mod_delete, attrs_dict)


def _dict_mod_add(to_dict, attrs_dict):
    """Adds attributes from attrs_dict to to_dict.

    :param to_dict: The dictionary to modify
    :type to_dict: dict(str, list[str]) or AttrsDict
    :param attrs_dict: Dictionary of attributes to add to ``to_dict``
    :type attrs_dict: dict(str, list[str]) or AttrsDict
    :return: None, modifies ``to_dict`` in place.
    :rtype: None
    """
    for attr, vals in six.iteritems(attrs_dict):
        if attr not in to_dict:
            to_dict[attr] = vals
        else:
            for val in vals:
                if val not in to_dict[attr]:
                    to_dict[attr].append(val)


def _dict_mod_replace(to_dict, attrs_dict):
    """Replaces attribute values in to_dict with those from attrs_dict.

    :param to_dict: The dictionary to modify
    :type to_dict: dict(str, list[str]) or AttrsDict
    :param attrs_dict: Dictionary of attributes to set in ``to_dict``
    :type attrs_dict: dict(str, list[str]) or AttrsDict
    :return: None, modifies ``to_dict`` in place.
    :rtype: None
    """
    to_dict.update(attrs_dict)


def _dict_mod_delete(to_dict, attrs_dict):
    """Deletes attribute values from to_dict that appear in attrs_dict

    :param to_dict: The dictionary to modify
    :type to_dict: dict(str, list[str]) or AttrsDict
    :param attrs_dict: Dictionary of attributes to delete from ``to_dict``. If any value is an empty list, delete all
                       attribute values.
    :type attrs_dict: dict(str, list[str]) or AttrsDict
    :return: None, modifies ``to_dict`` in place.
    :rtype: None
    """
    for attr, delVals in six.iteritems(attrs_dict):
        if attr in to_dict:
            if delVals:
                for val in delVals:
                    try:
                        to_dict[attr].remove(val)
                    except Exception:
                        pass
            else:
                del to_dict[attr]








from . import exceptions
import re

try:
    str.casefold

    def casefold(value):
        return value.casefold()
except AttributeError:
    from py2casefold import casefold


def re_anchor(r):
    return r'^' + r + r'$'


def find_closing_paren(text):
    if text[0] != '(':
        raise ValueError('missing opening paren')
    parens = 1
    i = 0
    try:
        while parens > 0:
            i += 1
            if text[i] == '(':
                parens += 1
            elif text[i] == ')':
                parens -= 1
    except IndexError:
        raise ValueError('missing closing paren')
    return i


def validate_phone_number(s):
    """Perform simplistic phone number validation"""

    # strip out non-digit and non-plus characters
    s = re.sub('[^0-9+]', '', s)

    # remove leading +
    if s.startswith('+'):
        s = s[1:]

    # Should only have numbers now
    if not s.isdigit():
        return False

    # Check length
    l = len(s)
    return (l >= 7 and l <= 15)


def collapse_whitespace(s):
    """Collapse all whitespace sequences down to a single space"""
    return re.sub('\s+', ' ', s).strip()


def _any_case(m):
    """For use with re.sub

     Convert a single-character lower-cased regex match to a character class
     with both cases
    """
    return '[{0}{1}]'.format(m.group(0), m.group(0).upper())


def escaped_regex(escape_chars):
    """
     Convert a series of characters to a regular expression that requires LDAP
     hex escapes or any single character excluding escape_chars

     Result will be a parenthesized subpattern that can be repeated to form a
     larger string.
    """

    # ensure no duplicate characters
    escape_chars = ''.join(set(escape_chars))

    try:
        # move dash to end for use in character class
        di = escape_chars.index('-')
        escape_chars = escape_chars[0:di] + escape_chars[di + 1:] + '-'
    except ValueError:
        # no dash
        pass

    # build hex-escape subpatterns
    subpatterns = []
    for c in escape_chars:
        code = hex(ord(c))
        if len(code) > 4:
            # only single-byte characters will ever need escaping in LDAP
            # multi-byte escapes probably not even supported
            raise ValueError('Character "{0}" does not need escaping'.format(c))

        # strip leading 0x
        code = code[2:]

        # zero-pad if necessary
        if len(code) == 1:
            code = '0' + code

        # make it an LDAP hex escape
        code = r'\\' + code

        # allow upper/lower cased alpha characters - hex() produces in lower
        code = re.sub('[a-f]', _any_case, code)

        subpatterns.append(code)

    # add final character class excluding escape_chars
    try:
        # double any backslash present
        bi = escape_chars.index('\\')
        escape_chars = escape_chars[0:bi] + '\\' + escape_chars[bi:]
    except ValueError:
        # no backslash
        pass
    subpatterns.append('[^{0}]'.format(escape_chars))

    return '({0})'.format('|'.join(subpatterns))


class CaseIgnoreDict(dict):
    """A dictionary with case-insensitive keys and storage of last actual key casing"""
    def __init__(self, plaindict=None):
        self._keys = {}
        if plaindict is not None:
            self.update(plaindict)

    def __setitem__(self, key, value):
        self._keys[key.lower()] = key
        dict.__setitem__(self, key, value)

    def setdefault(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            self[key] = default
            return default

    def __getitem__(self, key):
        key = self._keys[key.lower()]
        return dict.__getitem__(self, key)

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def __contains__(self, key):
        try:
            self[key]
            return True
        except KeyError:
            return False

    def update(self, other):
        for key in other:
            self[key] = other[key]

    def __delitem__(self, key):
        lkey = key.lower()
        key = self._keys[lkey]
        dict.__delitem__(self, key)
        del self._keys[lkey]

    def clear(self):
        dict.clear(self)
        self._keys.clear()


def get_one_result(results):
    n = len(results)
    if n == 0:
        raise exceptions.NoSearchResults()
    elif n > 1:
        raise exceptions.MultipleSearchResults()
    else:
        return results[0]


_get_class_module_err_msg = ('Could not identify the source module for object {0}. This may indicate an incompatability'
                             ' between your Python version or implementation and laurelin.')


def get_obj_module(obj):
    """Identify the name of the module where the given object was defined.

    This uses the __module__ attribute with some error handling since there is some suggestion around the internet
    that this attribute may not be 100% reliable. But, the consensus seems to be that it is at least "pretty reliable"
    so this function performs a little bit of error handling just in case. It could probably do more, but the actual
    presentation of lack of support for this is unknown.
    """
    try:
        modname = obj.__module__
        if not modname:
            raise exceptions.LDAPError(_get_class_module_err_msg.format(obj.__name__))
        return modname
    except AttributeError:
        raise exceptions.LDAPError(_get_class_module_err_msg.format(obj.__name__))



from __future__ import absolute_import
from . import rfc4512
from . import utils
from .attributetype import get_attribute_type
from .exceptions import LDAPSchemaError, LDAPWarning
from .protoutils import parse_qdescrs
from .utils import CaseIgnoreDict

import logging
import re
from warnings import warn

_re_object_class = re.compile(utils.re_anchor(rfc4512.ObjectClassDescription))

_oid_object_classes = {}
_name_object_classes = CaseIgnoreDict()

logger = logging.getLogger(__name__)


def get_object_class(ident):
    """Get an instance of :class:`ObjectClass` associated with either a name or an OID

    :param str ident: Either the numeric OID of the desired object class spec or one of its specified names
    :return: The ObjectClass associated with the name/OID
    :rtype: ObjectClass
    """
    if ident[0].isdigit():
        return _oid_object_classes[ident]
    else:
        try:
            return _name_object_classes[ident]
        except KeyError:
            return DefaultObjectClass(ident)


def _parse_oids(spec):
    if not spec:
        return []
    spec = spec.strip('() ')
    spec = spec.split('$')
    return [oid.strip() for oid in spec]


def _parse_attr_list(spec):
    ret = []
    for attr in _parse_oids(spec):
        if attr[0].isdigit():
            try:
                for name in get_attribute_type(attr).names:
                    ret.append(name)
            except KeyError:
                warn('Attribute type OID {0} has not been defined, excluding from '
                     'valid attributes on object class'.format(attr),
                     LDAPWarning)
        else:
            ret.append(attr)
    return ret


class ObjectClass(object):
    """Parses an LDAP object class specification and implements superclass inheritance.

    Each instantiation registers the names and OID specified so that they can later be access with
    :func:`get_object_class`.

    See the :mod:`laurelin.ldap.schema` module source for example usages.

    :param str spec: The LDAP specification for an object class
    :raises LDAPSchemaError:
         * if the schema is syntactically invalid
         * if the OID specified has already been registered
         * if one of the names specified has already been registered

    :var str oid: The specified OID
    :var tuple(str) names: All specified names
    :var list[str] superclasses: The list of all specified superclass names/OIDs.
    :var str kind: One of `ABSTRACT`, `STRUCTURAL`, or `AUXILIARY`
    :var bool obsolete: True if the objectClass has been marked obsolete.
    :var list[str] my_must: The list of required attribute types for this class
    :var list[str] my_may: The list of allowed attribute types for this class
    """
    def __init__(self, spec):
        spec = utils.collapse_whitespace(spec).strip()
        m = _re_object_class.match(spec)
        if not m:
            raise LDAPSchemaError('Invalid object class description')

        self.oid = m.group('oid')

        self.names = parse_qdescrs(m.group('name'))

        self.superclasses = _parse_oids(m.group('superclass'))

        kind = m.group('kind')
        if kind is not None:
            self.kind = kind
        else:
            self.kind = 'STRUCTURAL'

        obsolete = m.group('obsolete')
        if obsolete is not None:
            self.obsolete = bool(obsolete)
        else:
            self.obsolete = False

        self.my_must = _parse_attr_list(m.group('must'))
        self.my_may = _parse_attr_list(m.group('may'))

        self._must = None
        self._may = None

    def register(self):
        # register OID
        if self.oid in _oid_object_classes:
            raise LDAPSchemaError('Duplicate object class OID {0}'.format(self.oid))
        _oid_object_classes[self.oid] = self

        # register names
        for name in self.names:
            if name in _name_object_classes:
                raise LDAPSchemaError('Duplicate object class name {0}'.format(name))
            _name_object_classes[name] = self

    @property
    def must(self):
        """Obtains all required attribute types after ascending the superclass specifications"""
        if self._must is not None:
            return self._must
        elif self.superclasses:
            self._must = self.my_must
            for oc in self.superclasses:
                self._must += get_object_class(oc).must
            return self._must
        else:
            self._must = self.my_must
            return self._must

    @property
    def may(self):
        """Obtains all allowed attribute types after ascending the superclass specifications"""
        if self._may is not None:
            return self._may
        elif self.superclasses:
            self._may = self.my_may
            for oc in self.superclasses:
                self._may += get_object_class(oc).may
            return self._may
        else:
            self._may = self.my_may
            return self._may

    def allowed_attr(self, name):
        """Check if the given attribute type name is allowed.

        :param name: The name of the attribute type to check
        :return: True if the given attribute type is allowed.
        :rtype: bool
        """
        return (name in self.may or name in self.must)

    def required_attr(self, name):
        """Check if the given attribute type name is required.

        :param name: The name of the attribute type to check
        :return: True if the given attribute type is required.
        :rtype: bool
        """
        return (name in self.must)

    def __repr__(self):
        return '<{0} "{1}">'.format(self.__class__.__name__, self.names[0])


class DefaultObjectClass(ObjectClass):
    """The default ObjectClass returned by :func:`get_object_class` when the requested object class is undefined.

    Users should probably never instantiate this.
    """
    def __init__(self, name):
        logger.debug('Using DefaultObjectClass for name={0}'.format(name))
        self.oid = None
        self.names = (name,)
        self.supertype = None
        self.kind = 'STRUCTURAL'
        self.obsolete = False
        self.superclasses = ()
        self.my_must = []
        self.my_may = []
        self._must = None
        self._may = None


## RFC 4512 4.3 extensibleObject


class ExtensibleObjectClass(ObjectClass):
    """The `extensibleObject` auxiliary objectClass allows entries that belong to it to hold any user attribute."""
    def allowed_attr(self, name):
        return True



from .base import (
    add_extension,
    Extensible,
    ExtensibleClass,
    ExtensionsBase,
    EXTENSION_CLSNAME,
    CLASS_EXTENSION_FMT,
)
from .registration import LaurelinRegistrar, LaurelinTransiter
from .user_base import (
    BaseLaurelinExtension,
    BaseLaurelinLDAPExtension,
    BaseLaurelinLDAPObjectExtension,
    BaseLaurelinSchema,
    BaseLaurelinControls,
)
from .laurelin_extensions import Extensions

extensions = Extensions()

__all__ = [
    'add_extension',
    'LaurelinRegistrar',
    'LaurelinTransiter',
    'BaseLaurelinExtension',
    'BaseLaurelinLDAPExtension',
    'BaseLaurelinLDAPObjectExtension',
    'BaseLaurelinSchema',
    'BaseLaurelinControls',
    'Extensible',
    'ExtensibleClass',
    'ExtensionsBase',
    'EXTENSION_CLSNAME',
    'CLASS_EXTENSION_FMT',
    'extensions'
]


"""Base classes for syntax rules and matching rules"""

from __future__ import absolute_import
from .exceptions import InvalidSyntaxError, LDAPSchemaError
from .utils import CaseIgnoreDict, get_obj_module
import re

## Syntax Rules

_oid_syntax_rules = {}
_oid_syntax_rule_objects = {}


def get_syntax_rule(oid):
    obj = _oid_syntax_rule_objects.get(oid)
    if not obj:
        obj = _oid_syntax_rules[oid]()
    return obj


class SyntaxRule(object):
    """Base class for all syntax rules"""

    OID = ''
    """The globally unique numeric OID of the syntax rule. Referenced in attribute type and matching rule specs. Must be
    defined by subclasses."""

    DESC = ''
    """Short text description of the rule. Must be defined by subclasses."""

    def __init__(self):
        if self.OID:
            if self.OID in _oid_syntax_rule_objects:
                raise LDAPSchemaError('Multiple instantiations of syntax rule with OID {0}'.format(self.OID))
            _oid_syntax_rule_objects[self.OID] = self

    @classmethod
    def register(cls):
        if cls.OID in _oid_syntax_rules:
            cur_class = _oid_syntax_rules[cls.OID].__name__
            cur_mod = get_obj_module(cur_class)
            new_class = cls.__name__
            new_mod = get_obj_module(new_class)
            raise LDAPSchemaError('Duplicate OID {0} in syntax rule declaration (original class {1}.{2}, '
                                  'new class {3}.{4})'.format(cls.OID, cur_mod, cur_class, new_mod, new_class))
        _oid_syntax_rules[cls.OID] = cls

    def validate(self, s):
        """Validate a string. Must be implemented by subclasses.

        :param s: Candidate string
        :return: Any useful value for the rule
        :raises InvalidSyntaxError: if the string is invalid
        """
        raise NotImplementedError()


class RegexSyntaxRule(SyntaxRule):
    """For validating rules based on a regular expression. Most syntax rules can inherit from this."""

    regex = r''
    """The regular expression defining the rule. Subclasses must define this attribute."""

    def __init__(self):
        self.compiled_re = re.compile(self.regex)
        SyntaxRule.__init__(self)

    def validate(self, s):
        """Validate a string against the regular expression.

        :param s: Candidate string
        :return: The regex match object
        :rtype: MatchObject
        :raises InvalidSyntaxError: if the string does not match
        """
        m = self.compiled_re.match(s)
        if m:
            return m
        else:
            raise InvalidSyntaxError('Not a valid {0}: {1}'.format(self.DESC, s))


## Matching Rules


_oid_matching_rules = {}
_name_matching_rules = CaseIgnoreDict()
_oid_matching_rule_objects = {}
_name_matching_rule_objects = CaseIgnoreDict()


def get_matching_rule(ident):
    """Obtains matching rule instance for name or OID"""
    if ident[0].isdigit():
        cls_dict = _oid_matching_rules
        obj_dict = _oid_matching_rule_objects
    else:
        cls_dict = _name_matching_rules
        obj_dict = _name_matching_rule_objects
    obj = obj_dict.get(ident)
    if not obj:
        obj = cls_dict[ident]()
    return obj


class MatchingRule(object):
    """Base class for all matching rules"""

    OID = ''
    """Globally unique numeric OID for the matching rule. This must be defined by subclasses."""

    NAME = ''
    """Globally unique name for the matching rule. Most attribute type specs will reference rules using the name, but
    they can also use the OID. This must be defined by subclasses."""

    SYNTAX = ''
    """The numeric OID for the syntax rule that assertion values must comply with. Subclasses must define this."""

    prep_methods = ()
    """A tuple of callables used to prepare attribute and asserion values. Subclasses may optionally define this."""

    def __init__(self):
        if self.OID:
            if self.OID in _oid_matching_rule_objects:
                raise LDAPSchemaError('Multiple instantiations of matching rule with OID {0}'.format(self.OID))
            _oid_matching_rule_objects[self.OID] = self
        if self.NAME:
            if self.NAME in _name_matching_rule_objects:
                raise LDAPSchemaError('Multiple instantiations of matching rule with name {0}'.format(self.NAME))
            _name_matching_rule_objects[self.NAME] = self

    @classmethod
    def register(cls):
        if cls.OID:
            if cls.OID in _oid_matching_rules:
                raise LDAPSchemaError('Duplicate OID {0} in matching rule declaration'.format(cls.OID))
            _oid_matching_rules[cls.OID] = cls
        if cls.NAME:
            if cls.NAME in _name_matching_rules:
                raise LDAPSchemaError('Duplicate name {0} in matching rule declaration'.format(cls.NAME))
            _name_matching_rules[cls.NAME] = cls

    def validate(self, value):
        """Perform validation according to the matching rule's syntax"""
        return get_syntax_rule(self.SYNTAX).validate(value)

    def prepare(self, value):
        """Prepare a string for matching"""
        for method in getattr(self, 'prep_methods', ()):
            value = method(value)
        return value

    def do_match(self, attribute_value, assertion_value):
        """Perform the match operation"""
        raise NotImplementedError()

    def match(self, attribute_value, assertion_value):
        """Prepare values and perform the match operation. Assumes values have
        already been validated.
        """
        attribute_value = self.prepare(attribute_value)
        assertion_value = self.prepare(assertion_value)
        return self.do_match(attribute_value, assertion_value)


# Note: currently only implementing equality matching rules since there is no
# use for ordering or substring matching rules for correct functioning of the
# current codebase. Any forseeable use for other types of rules would be in
# support of new features not currently planned.

class EqualityMatchingRule(MatchingRule):
    """Base class for all EQUALITY matching rules"""

    def do_match(self, attribute_value, assertion_value):
        """Perform equality matching"""
        return (attribute_value == assertion_value)



"""Provides protocol-level interface for low-level sockets"""

from __future__ import absolute_import
import six
import ssl
import logging
from glob import glob
from socket import create_connection, socket, error as SocketError
from six.moves.urllib.parse import unquote
from collections import deque
from puresasl.client import SASLClient

from .rfc4511 import LDAPMessage, ResultCode
from .exceptions import LDAPError, LDAPSASLError, LDAPConnectionError, LDAPUnsolicitedMessage, UnexpectedResponseType
from .protoutils import pack, unpack
from .pyasn1.codec.ber.encoder import encode as ber_encode
from .pyasn1.codec.ber.decoder import decode as ber_decode
from .pyasn1.error import SubstrateUnderrunError

try:
    from socket import AF_UNIX
    _have_unix_socket = True
except ImportError:
    AF_UNIX = None
    _have_unix_socket = False

_next_sock_id = 0
logger = logging.getLogger(__name__)


class LDAPSocket(object):
    """Holds a connection to an LDAP server.

    :param str host_uri: "scheme://netloc" to connect to
    :param int connect_timeout: Number of seconds to wait for connection to be accepted
    :param bool ssl_verify: Validate the certificate and hostname on an SSL/TLS connection
    :param str ssl_ca_file: Path to PEM-formatted concatenated CA certficates file
    :param str ssl_ca_path: Path to directory with CA certs under hashed file names. See
                            https://www.openssl.org/docs/man1.1.0/ssl/SSL_CTX_load_verify_locations.html for more
                            information about the format of this directory.
    :param ssl_ca_data: An ASCII string of one or more PEM-encoded certs or a bytes object containing DER-encoded
                        certificates.
    :type ssl_ca_data: str or bytes
    """

    RECV_BUFFER = 4096

    # For ldapi:/// try to connect to these socket files in order
    # Globs must match exactly one result
    LDAPI_SOCKET_PATHS = ['/var/run/ldapi', '/var/run/slapd/ldapi', '/var/run/slapd-*.socket']

    # OIDs of unsolicited messages

    OID_DISCONNECTION_NOTICE = '1.3.6.1.4.1.1466.20036'  # RFC 4511 sec 4.4.1 Notice of Disconnection

    def __init__(self, host_uri, connect_timeout=5, ssl_verify=True, ssl_ca_file=None, ssl_ca_path=None,
                 ssl_ca_data=None):

        self._prop_init(connect_timeout)
        self._uri_connect(host_uri, ssl_verify, ssl_ca_file, ssl_ca_path, ssl_ca_data)

    def _prop_init(self, connect_timeout=5):
        # get socket ID number
        global _next_sock_id
        self.ID = _next_sock_id
        _next_sock_id += 1

        # misc init
        self._message_queues = {}
        self._next_message_id = 1
        self._sasl_client = None

        self.refcount = 0
        self.bound = False
        self.unbound = False
        self.abandoned_mids = []
        self.started_tls = False
        self.connect_timeout = connect_timeout

    def _parse_uri(self, host_uri):
        # parse host_uri
        parts = host_uri.split('://')
        if len(parts) == 1:
            netloc = unquote(parts[0])
            if netloc[0] == '/':
                scheme = 'ldapi'
            else:
                scheme = 'ldap'
        elif len(parts) == 2:
            scheme = parts[0]
            netloc = unquote(parts[1])
        else:
            raise LDAPError('Invalid host_uri')
        self.uri = '{0}://{1}'.format(scheme, netloc)
        return scheme, netloc

    def _uri_connect(self, host_uri, ssl_verify, ssl_ca_file, ssl_ca_path, ssl_ca_data):
        # connect
        scheme, netloc = self._parse_uri(host_uri)
        logger.info('Connecting to {0} on #{1}'.format(self.uri, self.ID))
        if scheme == 'ldap':
            self._inet_connect(netloc, 389)
        elif scheme == 'ldaps':
            self._inet_connect(netloc, 636)
            self.start_tls(ssl_verify, ssl_ca_file, ssl_ca_path, ssl_ca_data)
            logger.info('Connected with TLS on #{0}'.format(self.ID))
        elif scheme == 'ldapi':
            if not _have_unix_socket:
                raise LDAPError('Unix sockets are not supported on your platform, please choose a protocol other'
                                'than ldapi')
            self.sock_path = None
            self._sock = socket(AF_UNIX)
            self.host = 'localhost'

            if netloc == '/':
                for sockGlob in LDAPSocket.LDAPI_SOCKET_PATHS:
                    fn = glob(sockGlob)
                    if not fn:
                        continue
                    if len(fn) > 1:
                        logger.debug('Multiple results for glob {0}'.format(sockGlob))
                        continue
                    fn = fn[0]
                    try:
                        self._connect(fn)
                        self.sock_path = fn
                        break
                    except SocketError:
                        continue
                if self.sock_path is None:
                    raise LDAPConnectionError('Could not find any local LDAPI unix socket - full '
                                              'socket path must be supplied in URI')
            else:
                try:
                    self._connect(netloc)
                    self.sock_path = netloc
                except SocketError as e:
                    raise LDAPConnectionError('failed connect to unix socket {0} - {1} ({2})'.format(
                        netloc, e.strerror, e.errno
                    ))

            logger.debug('Connected to unix socket {0} on #{1}'.format(self.sock_path, self.ID))
        else:
            raise LDAPError('Unsupported scheme "{0}"'.format(scheme))

    def _connect(self, addr):
        self._sock.settimeout(self.connect_timeout)
        self._sock.connect(addr)
        self._sock.settimeout(None)

    def _inet_connect(self, netloc, default_port):
        ap = netloc.rsplit(':', 1)
        self.host = ap[0]
        if len(ap) == 1:
            port = default_port
        else:
            port = int(ap[1])
        try:
            self._sock = create_connection((self.host, port), self.connect_timeout)
            logger.debug('Connected to {0}:{1} on #{2}'.format(self.host, port, self.ID))
        except SocketError as e:
            raise LDAPConnectionError('failed connect to {0}:{1} - {2} ({3})'.format(
                                      self.host, port, e.strerror, e.errno))

    def start_tls(self, verify=True, ca_file=None, ca_path=None, ca_data=None):
        """Install TLS layer on this socket connection.

        :param bool verify: Validate the certificate and hostname on an SSL/TLS connection
        :param str ca_file: Path to PEM-formatted concatenated CA certficates file
        :param str ca_path: Path to directory with CA certs under hashed file names. See
                            https://www.openssl.org/docs/man1.1.0/ssl/SSL_CTX_load_verify_locations.html for more
                            information about the format of this directory.
        :param ca_data: An ASCII string of one or more PEM-encoded certs or a bytes object containing DER-encoded
                        certificates.
        :type ca_data: str or bytes
        """
        if self.started_tls:
            raise LDAPError('TLS layer already installed')

        if verify:
            verify_mode = ssl.CERT_REQUIRED
        else:
            verify_mode = ssl.CERT_NONE

        try:
            proto = ssl.PROTOCOL_TLS
        except AttributeError:
            proto = ssl.PROTOCOL_SSLv23

        try:
            ctx = ssl.SSLContext(proto)
            ctx.verify_mode = verify_mode
            ctx.check_hostname = False  # we do this ourselves
            if verify:
                ctx.load_default_certs()
            if ca_file or ca_path or ca_data:
                ctx.load_verify_locations(cafile=ca_file, capath=ca_path, cadata=ca_data)
            self._sock = ctx.wrap_socket(self._sock)
        except AttributeError:
            # SSLContext wasn't added until 2.7.9
            if ca_path or ca_data:
                raise RuntimeError('python version >= 2.7.9 required for SSL ca_path/ca_data')

            self._sock = ssl.wrap_socket(self._sock, ca_certs=ca_file, cert_reqs=verify_mode, ssl_version=proto)

        if verify:
            cert = self._sock.getpeercert()
            cert_cn = dict([e[0] for e in cert['subject']])['commonName']
            self.check_hostname(cert_cn, cert)
        else:
            logger.debug('Skipping hostname validation')
        self.started_tls = True
        logger.debug('Installed TLS layer on #{0}'.format(self.ID))

    def check_hostname(self, cert_cn, cert):
        """SSL check_hostname according to RFC 4513 sec 3.1.3. Compares supplied values against ``self.host`` to
        determine the validity of the cert.

        :param str cert_cn: The common name of the cert
        :param dict cert: A dictionary representing the rest of the cert. Checks key subjectAltNames for a list of
                          (type, value) tuples, where type is 'DNS' or 'IP'. DNS supports leading wildcard.
        :rtype: None
        :raises LDAPConnectionError: if no supplied values match ``self.host``
        """
        if self.host == cert_cn:
            logger.debug('Matched server identity to cert commonName')
        else:
            valid = False
            tried = [cert_cn]
            for type, value in cert.get('subjectAltName', []):
                if type == 'DNS' and value.startswith('*.'):
                    valid = self.host.endswith(value[1:])
                else:
                    valid = (self.host == value)
                tried.append(value)
                if valid:
                    logger.debug('Matched server identity to cert {0} subjectAltName'.format(type))
                    break
            if not valid:
                raise LDAPConnectionError('Server identity "{0}" does not match any cert names: {1}'.format(
                    self.host, ', '.join(tried)))

    def sasl_init(self, mechs, **props):
        """Initialize a :class:`.puresasl.client.SASLClient`"""
        self._sasl_client = SASLClient(self.host, 'ldap', **props)
        self._sasl_client.choose_mechanism(mechs)

    def _has_sasl_client(self):
        return self._sasl_client is not None

    def _require_sasl_client(self):
        if not self._has_sasl_client():
            raise LDAPSASLError('SASL init not complete')

    @property
    def sasl_qop(self):
        """Obtain the chosen quality of protection"""
        self._require_sasl_client()
        return self._sasl_client.qop

    @property
    def sasl_mech(self):
        """Obtain the chosen mechanism"""
        self._require_sasl_client()
        mech = self._sasl_client.mechanism
        if mech is None:
            raise LDAPSASLError('SASL init not complete - no mech chosen')
        else:
            return mech

    def sasl_process_auth_challenge(self, challenge):
        """Process an auth challenge and return the correct response"""
        self._require_sasl_client()
        return self._sasl_client.process(challenge)

    def _prep_message(self, op, obj, controls=None):
        """Prepare a message for transmission"""
        mid = self._next_message_id
        self._next_message_id += 1
        lm = pack(mid, op, obj, controls)
        raw = ber_encode(lm)
        if self._has_sasl_client():
            raw = self._sasl_client.wrap(raw)
        return mid, raw

    def send_message(self, op, obj, controls=None):
        """Create and send an LDAPMessage given an operation name and a corresponding object.

        Operation names must be defined as component names in laurelin.ldap.rfc4511.ProtocolOp and
        the object must be of the corresponding type.

        :param str op: The protocol operation name
        :param object obj: The associated protocol object (see :class:`.rfc4511.ProtocolOp` for mapping.
        :param controls: Any request controls for the message
        :type controls: rfc4511.Controls or None
        :return: The message ID for this message
        :rtype: int
        """
        mid, raw = self._prep_message(op, obj, controls)
        self._sock.sendall(raw)
        return mid

    def recv_one(self, want_message_id):
        """Get the next message with ``want_message_id`` being sent by the server

        :param int want_message_id: The desired message ID.
        :return: The LDAP message
        :rtype: rfc4511.LDAPMessage
        """
        return next(self.recv_messages(want_message_id))

    def recv_messages(self, want_message_id):
        """Iterate all messages with ``want_message_id`` being sent by the server.

        :param int want_message_id: The desired message ID.
        :return: An iterator over :class:`.rfc4511.LDAPMessage`.
        """
        flush_queue = True
        raw = b''
        while True:
            if flush_queue:
                if want_message_id in self._message_queues:
                    q = self._message_queues[want_message_id]
                    while True:
                        if len(q) == 0:
                            break
                        obj = q.popleft()
                        if len(q) == 0:
                            del self._message_queues[want_message_id]
                        yield obj
            else:
                flush_queue = True
            if want_message_id in self.abandoned_mids:
                return
            try:
                newraw = self._sock.recv(LDAPSocket.RECV_BUFFER)
                if self._has_sasl_client():
                    newraw = self._sasl_client.unwrap(newraw)
                raw += newraw
                while len(raw) > 0:
                    response, raw = ber_decode(raw, asn1Spec=LDAPMessage())
                    have_message_id = response.getComponentByName('messageID')
                    if want_message_id == have_message_id:
                        yield response
                    elif have_message_id == 0:
                        msg = 'Received unsolicited message (default message - should never be seen)'
                        try:
                            mid, xr, ctrls = unpack('extendedResp', response)
                            res_code = xr.getComponentByName('resultCode')
                            xr_oid = six.text_type(xr.getComponentByName('responseName'))
                            if xr_oid == LDAPSocket.OID_DISCONNECTION_NOTICE:
                                mtype = 'Notice of Disconnection'
                            else:
                                mtype = 'Unhandled ({0})'.format(xr_oid)
                            diag = xr.getComponentByName('diagnosticMessage')
                            msg = 'Got unsolicited message: {0}: {1}: {2}'.format(mtype, res_code, diag)
                            if res_code == ResultCode('protocolError'):
                                msg += (' (This may indicate an incompatability between laurelin-ldap and your server '
                                        'distribution)')
                            elif res_code == ResultCode('strongerAuthRequired'):
                                # this is a direct quote from RFC 4511 sec 4.4.1
                                msg += (' (The server has detected that an established security association between the'
                                        ' client and server has unexpectedly failed or been compromised)')
                        except UnexpectedResponseType:
                            msg = 'Unhandled unsolicited message from server'
                        finally:
                            raise LDAPUnsolicitedMessage(response, msg)
                    else:
                        if have_message_id not in self._message_queues:
                            self._message_queues[have_message_id] = deque()
                        self._message_queues[have_message_id].append(response)
            except SubstrateUnderrunError:
                flush_queue = False
                continue

    def close(self):
        """Close the low-level socket connection."""
        return self._sock.close()



"""Extension adding netgroup support to laurelin.

Includes schema definitions.

You should begin by tagging the base object which all netgroups are below, and defining the RDN attribute and scope. If
the structure is flat there is a performance advantage by setting ``relative_search_scope=Scope.ONE``::

    from laurelin.ldap import LDAP, Scope, extensions

    with LDAP() as ldap:
        ldap.base.obj('ou=netgroups',
                      tag=extensions.netgroups.TAG,
                      relative_search_scope=Scope.ONE,
                      rdn_attr='cn')

Member Lists
^^^^^^^^^^^^

This extension module allows a shortcut to specify members of netgroups. Any function with a ``members`` argument uses
this feature.

The function name will tell you whether it expects users (e.g., ``add_users``) or hosts (e.g. ``add_hosts``). If you
just specify a string in your member list, it will be assumed to be either a user or a host accordingly.

You can also specify a tuple with up to 3 elements for any member list entry. These fields must correspond to the
``nisNetgroupTriple`` fields: host, user, and domain. For user functions, at least the first 2 tuple elements must be
specified. For host functions, only the first is required, the 2nd (user) field will be assumed as an empty string. In
all cases, the domain can be specified for all members by passing the ``domain`` argument to the function (it defaults
to an empty string).

The third option for member list entries is to specify the full ``nisNetgroupTriple`` yourself in a string.

Finally, you can specify a ``memberNisNetgroup`` by prefixing the entry with a ``+`` symbol. For example: ``+users``.

Examples::

    users = [
       'alice',
       'bob',
       ('dir.example.org', 'admin'),
       '(dir.example.org,manager,secrets.example.org)',
    ]

    ldap.netgroups.add_users('cn=managers,ou=netgroups,dc=example,dc=org', users, domain='example.org')
    # Adds the following nisNetgroupTriples:
    #  (,alice,example.org)
    #  (,bob,example.org)
    #  (dir.example.org,admin,example.org)
    #  (dir.example.org,manager,secrets.example.org)
    # Does not add any memberNisNetgroups

    hosts = [
        'dir1.example.org',
        'dir2.example.org',
        '(dir3.example.org,,)',
        ('dir4.example.org',),
        '+aws_backup_dir_servers',
    ]

    ldap.netgroups.add_hosts('cn=dir_servers,ou=netgroups,dc=example,dc=org', hosts)
    # Adds the following nisNetgroupTriples:
    #  (dir1.example.org,,)
    #  (dir2.example.org,,)
    #  (dir3.example.org,,)
    #  (dir4.example.org,,)
    # Adds the following memberNisNetgroup:
    #  aws_backup_dir_servers
"""
from __future__ import absolute_import
import re
from laurelin.ldap import (
    AttributeType,
    ObjectClass,
    get_object_class,
    RegexSyntaxRule,
    LDAPError,
    extensions,
    BaseLaurelinExtension,
    BaseLaurelinSchema,
    BaseLaurelinLDAPExtension,
    BaseLaurelinLDAPObjectExtension,
)
import six

extensions.base_schema.require()

TAG = 'netgroup_base'

_TRIPLE_RE = '^\(([^,]*),([^,]*),([^)]*)\)$'


class LaurelinExtension(BaseLaurelinExtension):
    NAME = 'netgroups'
    TAG = TAG


class LaurelinSchema(BaseLaurelinSchema):
    NIS_NETGROUP = ObjectClass('''
    ( 1.3.6.1.1.1.2.8 NAME 'nisNetgroup' SUP top STRUCTURAL
      MUST cn
      MAY ( nisNetgroupTriple $ memberNisNetgroup $ description ) )
    ''')

    MEMBER_NIS_NETGROUP = AttributeType('''
    ( 1.3.6.1.1.1.1.13 NAME 'memberNisNetgroup'
      EQUALITY caseExactIA5Match
      SUBSTR caseExactIA5SubstringsMatch
      SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 )
    ''')

    NIS_NETGROUP_TRIPLE = AttributeType('''
    ( 1.3.6.1.1.1.1.14 NAME 'nisNetgroupTriple'
      DESC 'Netgroup triple'
      EQUALITY caseExactMatch
      SYNTAX 1.3.6.1.1.1.0.0 )
    ''')

    class NisNetgroupTripleSytnax(RegexSyntaxRule):
        OID = '1.3.6.1.1.1.0.0'
        DESC = 'NIS netgroup triple'
        regex = _TRIPLE_RE


OBJECT_CLASS = 'nisNetgroup'


def _netgroup_attrs():
    oc = get_object_class(OBJECT_CLASS)
    return oc.must + oc.may


def _netgroup_attrs_arg(attrs_arg):
    if attrs_arg is None:
        return _netgroup_attrs()
    else:
        return attrs_arg


class LaurelinLDAPExtension(BaseLaurelinLDAPExtension):
    def get(self, cn, attrs=None):
        """Find a specific netgroup object.

        This depends on the base object having been tagged and configured properly. See
        :mod:`laurelin.extensions.netgroups`.

        :param str cn: The name of the group or an RDN
        :param list[str] attrs: List of attribute names to get. Defaults to all netgroup attributes.
        :return: The netgroup object
        :rtype: LDAPObject
        :raises TagError: if the base object has not been tagged.
        """
        attrs = _netgroup_attrs_arg(attrs)
        return self.parent.tag(TAG).find(cn, attrs)

    def search(self, filter, attrs=None):
        """Search for netgroups.

        This depends on the base object having been tagged and configured properly. See
        :mod:`laurelin.extensions.netgroups`.

        :param str filter: A partial filter string. The nisNetgroup objectClass will automatically be included in the
                           filter sent to the server.
        :param list[str] attrs: List of attribute names to get. Defaults to all netgroup attributes.
        :return: An iterator over matching netgroup objects, yielding instances of :class:`.LDAPObject`.
        :rtype: SearchResultHandle
        :raises TagError: if the base object has not been tagged.
        """
        attrs = _netgroup_attrs_arg(attrs)
        return self.parent.tag(TAG).search(_netgroup_filter(filter), attrs)

    def get_obj_users(self, ng_obj, recursive=True):
        """Get a list of netgroup users from an already queried object, possibly querying for memberNisNetgroups if
        ``recursive=True`` (the default).

        :param LDAPObject ng_obj: A netgroup LDAP object
        :param bool recursive: Set to False to only consider members of this group directly
        :return: A list of usernames
        :rtype: list[str]
        """
        users = _extract_triple_field(ng_obj, 2)
        if recursive and ('memberNisNetgroup' in ng_obj):
            for member in ng_obj['memberNisNetgroup']:
                users += self.get_users(member, True)
        return users

    def get_users(self, cn, recursive=True):
        """Get a list of all user entries for a netgroup.

        This depends on the base object having been tagged and configured properly. See
        :mod:`laurelin.extensions.netgroups`.

        :param str cn: The name of the group or an RDN
        :param bool recursive: Recursively get users by following memberNisNetgroups
        :return: A list of usernames
        :rtype: list[str]
        :raises TagError: if the base object has not been tagged.
        """
        ng = self.get(cn)
        return self.get_obj_users(ng, recursive)

    def get_obj_hosts(self, ng_obj, recursive=True):
        """Get a list of netgroup hosts from an already queried object, possibly querying for memberNisNetgroups if
        ``recursive=True`` (the default).

        :param LDAPObject ng_obj: A netgroup LDAP object
        :param bool recursive: Set to False to only consider members of this group directly
        :return: A list of hostnames
        :rtype: list[str]
        """
        hosts = _extract_triple_field(ng_obj, 1)
        if recursive and ('memberNisNetgroup' in ng_obj):
            for member in ng_obj['memberNisNetgroup']:
                hosts += self.get_hosts(member, True)
        return hosts

    def get_hosts(self, cn, recursive=True):
        """Query a list of all host entries for a netgroup.

        This depends on the base object having been tagged and configured properly. See
        :mod:`laurelin.extensions.netgroups`.

        :param str cn: The name of the group or an RDN
        :param bool recursive: Recursively get hosts by following memberNisNetgroups
        :return: A list of hostnames
        :rtype: list[str]
        :raises TagError: if the base object has not been tagged.
        """
        ng = self.get(cn)
        return self.get_obj_hosts(ng, recursive)

    def add_users(self, dn, members, domain=''):
        """Add new users to a netgroup.

        :param str dn: The absolute DN of the netgroup object
        :param members: A Member List (see :mod:`laurelin.extensions.netgroups` doc) or sinlge member list entry
        :type members: list or str or tuple
        :param str domain: The default domain to use in nisNetgroupTriples where not already specified
        :rtype: None
        """
        self.parent.add_attrs(dn, _member_user_list_to_attrs(members, domain))

    def add_hosts(self, dn, members, domain=''):
        """Add new hosts to a netgroup.

        :param str dn: The absolute DN of the netgroup object
        :param members: A Member List (see :mod:`laurelin.extensions.netgroups` doc) or single member list entry
        :type members: list or str or tuple
        :param str domain: The default domain to use in nisNetgroupTriples where not already specified
        :rtype: None
        """
        self.parent.add_attrs(dn, _member_host_list_to_attrs(members, domain))

    def replace_users(self, dn, members, domain=''):
        """Set new users on a netgroup.

        :param str dn: The absolute DN of the netgroup object
        :param members: A Member List (see :mod:`laurelin.extensions.netgroups` doc) or sinlge member list entry
        :type members: list or str or tuple
        :param str domain: The default domain to use in nisNetgroupTriples where not already specified
        :rtype: None
        """
        self.parent.replace_attrs(dn, _member_user_list_to_attrs(members, domain))

    def replace_hosts(self, dn, members, domain=''):
        """Set new hosts on a netgroup.

        :param str dn: The absolute DN of the netgroup object
        :param members: A Member List (see :mod:`laurelin.extensions.netgroups` doc) or single member list entry
        :type members: list or str or tuple
        :param str domain: The default domain to use in nisNetgroupTriples where not already specified
        :rtype: None
        """
        self.parent.replace_attrs(dn, _member_host_list_to_attrs(members, domain))

    def delete_users(self, dn, members, domain=''):
        """Delete users from a netgroup.

        :param str dn: The absolute DN of the netgroup object
        :param members: A Member List (see :mod:`laurelin.extensions.netgroups` doc) or sinlge member list entry
        :type members: list or str or tuple
        :param str domain: The default domain to use in nisNetgroupTriples where not already specified
        :rtype: None
        """
        self.parent.delete_attrs(dn, _member_user_list_to_attrs(members, domain))

    def delete_hosts(self, dn, members, domain=''):
        """Delete hosts from a netgroup.

        :param str dn: The absolute DN of the netgroup object
        :param members: A Member List (see :mod:`laurelin.extensions.netgroups` doc) or single member list entry
        :type members: list or str or tuple
        :param str domain: The default domain to use in nisNetgroupTriples where not already specified
        :rtype: None
        """
        self.parent.delete_attrs(dn, _member_host_list_to_attrs(members, domain))


class LaurelinLDAPObjectExtension(BaseLaurelinLDAPObjectExtension):
    def _require_netgroup(self):
        """Requires that this :class:`.LDAPObject` has the required netgroup object class.

        :raises RuntimeError: if the object is missing the netgroup object class
        """
        if not self.parent.has_object_class(OBJECT_CLASS):
            raise RuntimeError('objectClass {0} is required'.format(OBJECT_CLASS))

    def get_users(self, recursive=True):
        """Get all users in this netgroup object.

        :param bool recursive: Set to False to ignore any memberNisNetgroups defined for this object.
        :return: A list of usernames
        :rtype: list[str]
        :raises RuntimeError: if this object is missing the netgroup object class
        """
        self._require_netgroup()
        return self.parent.ldap_conn.netgroups.get_obj_users(self, recursive)

    def get_hosts(self, recursive=True):
        """Get all hosts in this netgroup object.

        :param bool recursive: Set to False to ignore any memberNisNetgroups defined for this object.
        :return: A list of hostnames
        :rtype: list[str]
        :raises RuntimeError: if this object is missing the netgroup object class
        """
        self._require_netgroup()
        return self.parent.ldap_conn.netgroups.get_obj_hosts(self, recursive)

    def add_users(self, members, domain=''):
        """Add new user netgroup entries to this netgroup object.

        :param members: A Member List (see :mod:`laurelin.extensions.netgroups` doc) or single member list entry
        :type members: list or str or tuple
        :param str domain: The default domain to use in nisNetgroupTriples where not already specified
        :rtype: None
        :raises RuntimeError: if this object is missing the netgroup object class
        """
        self._require_netgroup()
        self.parent.add_attrs(_member_user_list_to_attrs(members, domain))

    def add_hosts(self, members, domain=''):
        """Add new host netgroup entries to this netgroup object.

        :param members: A Member List (see :mod:`laurelin.extensions.netgroups` doc) or single member list entry
        :type members: list or str or tuple
        :param str domain: The default domain to use in nisNetgroupTriples where not already specified
        :rtype: None
        :raises RuntimeError: if this object is missing the netgroup object class
        """
        self._require_netgroup()
        self.parent.add_attrs(_member_host_list_to_attrs(members, domain))

    def replace_users(self, members, domain=''):
        """Set new user netgroup entries on this netgroup object.

        :param members: A Member List (see :mod:`laurelin.extensions.netgroups` doc) or single member list entry
        :type members: list or str or tuple
        :param str domain: The default domain to use in nisNetgroupTriples where not already specified
        :rtype: None
        :raises RuntimeError: if this object is missing the netgroup object class
        """
        self._require_netgroup()
        self.parent.replace_attrs(_member_user_list_to_attrs(members, domain))

    def replace_hosts(self, members, domain=''):
        """Set new host netgroup entries on this netgroup object.

        :param members: A Member List (see :mod:`laurelin.extensions.netgroups` doc) or single member list entry
        :type members: list or str or tuple
        :param str domain: The default domain to use in nisNetgroupTriples where not already specified
        :rtype: None
        :raises RuntimeError: if this object is missing the netgroup object class
        """
        self._require_netgroup()
        self.parent.replace_attrs(_member_host_list_to_attrs(members, domain))

    def delete_users(self, members, domain=''):
        """Delete user netgroup entries from this netgroup object.

        :param members: A Member List (see :mod:`laurelin.extensions.netgroups` doc) or single member list entry
        :type members: list or str or tuple
        :param str domain: The default domain to use in nisNetgroupTriples where not already specified
        :rtype: None
        :raises RuntimeError: if this object is missing the netgroup object class
        """
        self._require_netgroup()
        self.parent.delete_attrs(_member_user_list_to_attrs(members, domain))

    def delete_hosts(self, members, domain=''):
        """Delete host netgroup entries from this netgroup object.

        :param members: A Member List (see :mod:`laurelin.extensions.netgroups` doc) or single member list entry
        :type members: list or str or tuple
        :param str domain: The default domain to use in nisNetgroupTriples where not already specified
        :rtype: None
        :raises RuntimeError: if this object is missing the netgroup object class
        """
        self._require_netgroup()
        self.parent.delete_attrs(_member_host_list_to_attrs(members, domain))


## private functions


TRIPLE_RE = re.compile(_TRIPLE_RE)


def _netgroup_filter(filter):
    return '(&(objectClass={0}){1})'.format(OBJECT_CLASS, filter)


def _is_triple(val):
    return TRIPLE_RE.match(val) is not None


def _nis_netgroup_triple(host, user, domain):
    if not host:
        host = ''
    if not user:
        user = ''
    if not domain:
        domain = ''
    return '({0},{1},{2})'.format(host, user, domain)


def _extract_triple_field(ng_obj, index):
    ret = []
    for triple in ng_obj.get('nisNetgroupTriple', []):
        m = TRIPLE_RE.match(triple)
        if m is None:
            raise LDAPError('Invalid nisNetgroupTriple: {0}'.format(triple))
        else:
            ret.append(m.group(index))
    return ret


def _member_user_list_to_attrs(member_list, domain=''):
    if not isinstance(member_list, list):
        member_list = [member_list]
    attrs = {}
    for member in member_list:
        attr = 'nisNetgroupTriple'
        if isinstance(member, six.string_types):
            if member[0] == '+':
                attr = 'memberNisNetgroup'
                member = member[1:]
            else:
                if not _is_triple(member):
                    member = _nis_netgroup_triple('', member, domain)
        elif isinstance(member, tuple):
            if len(member) == 1:
                raise ValueError('At least first 2 triple values (host,user) must be specified for users')
            elif len(member) == 2:
                member = _nis_netgroup_triple(member[0], member[1], domain)
            elif len(member) == 3:
                member = _nis_netgroup_triple(*member)
            else:
                raise ValueError('tuple must have 2 or 3 elements')
        else:
            raise TypeError('member_list elements must be string or tuple')
        if attr not in attrs:
            attrs[attr] = []
        attrs[attr].append(member)
    return attrs


def _member_host_list_to_attrs(member_list, domain=''):
    if not isinstance(member_list, list):
        member_list = [member_list]
    attrs = {}
    for member in member_list:
        attr = 'nisNetgroupTriple'
        if isinstance(member, six.string_types):
            if member[0] == '+':
                attr = 'memberNisNetgroup'
                member = member[1:]
            else:
                if not _is_triple(member):
                    member = _nis_netgroup_triple(member, '', domain)
        elif isinstance(member, tuple):
            if len(member) == 1:
                member = _nis_netgroup_triple(member[0], '', domain)
            elif len(member) == 2:
                member = _nis_netgroup_triple(member[0], member[1], domain)
            elif len(member) == 3:
                member = _nis_netgroup_triple(*member)
            else:
                raise ValueError('tuple must have 1-3 elements')
        else:
            raise TypeError('member_list elements must be string or tuple')
        if attr not in attrs:
            attrs[attr] = []
        attrs[attr].append(member)
    return attrs



from .registration import LaurelinRegistrar, LaurelinTransiter


class BaseLaurelinExtension(LaurelinRegistrar, LaurelinTransiter):
    """Base class for basic extension class. Can house schema and controls definitions."""
    NAME = '__undefined__'
    INSTANCE = None


class BaseLaurelinLDAPExtension(object):
    """Base class for extensions to the LDAP class"""
    def __init__(self, parent):
        """
        :param laurelin.ldap.LDAP parent: The parent LDAP instance
        """
        self.parent = parent


class BaseLaurelinLDAPObjectExtension(object):
    """Base class for extensions to the LDAPObject class"""
    def __init__(self, parent):
        """
        :param laurelin.ldap.LDAPObject parent: The parent LDAPObject instance
        """
        self.parent = parent


class BaseLaurelinSchema(LaurelinTransiter):
    """Optional base class for a class defining schema elements - only subclassing LaurelinTransiter is required"""
    pass


class BaseLaurelinControls(LaurelinTransiter):
    """Optional base class for a class defining controls - only subclassing LaurelinTransiter is required"""
    pass



"""Contains utilities for handling filters.

See RFC4515 String Representation of Search Filters
"""

from __future__ import absolute_import
from parsimonious.grammar import Grammar
from parsimonious.exceptions import ParseError
import six
from six.moves import range

from . import rfc4511
from .exceptions import LDAPError

escape_map = [
    ('(', '\\28'),
    (')', '\\29'),
    ('&', '\\26'),
    ('|', '\\7c'),
    ('!', '\\21'),
    ('=', '\\3d'),
    ('<', '\\3c'),
    ('>', '\\3e'),
    ('~', '\\7e'),
    ('*', '\\2a'),
    ('/', '\\2f')
]


def escape(text):
    """Escape special characters"""
    for rep in escape_map:
        text = text.replace(*rep)
    return text


ava_grammar = '''
      rfc4515_ava    = substring / simple / extensible
      simple         = attr filtertype assertionvalue
      filtertype     = approx / greaterorequal / lessorequal / equal
      equal          = EQUALS
      approx         = TILDE EQUALS
      greaterorequal = RANGLE EQUALS
      lessorequal    = LANGLE EQUALS
      extensible     = ( ( attr dnattrs? matchingrule? COLON EQUALS assertionvalue )
                       / ( dnattrs? matchingrule COLON EQUALS assertionvalue ) )
      substring      = attr EQUALS initial? any final?
      initial        = assertionvalue
      any            = ASTERISK (assertionvalue ASTERISK)*
      final          = assertionvalue
      attr           = attributedescription

      attributedescription = attributetype options
      attributetype        = oid
      options              = ( SEMI option )*
      option               = keychar+

      dnattrs        = COLON "dn"
      matchingrule   = COLON oid

      oid = descr / numericoid

      numericoid = number ( DOT number )+
      number     = DIGIT / ( LDIGIT DIGIT+ )
      DIGIT      = ~r"[0-9]"
      LDIGIT     = ~r"[1-9]"

      descr       = keystring
      keystring   = leadkeychar keychar*
      leadkeychar = ALPHA
      keychar     = ALPHA / DIGIT / HYPHEN
      ALPHA       = ~r"[A-Za-z]"

      assertionvalue = valueencoding
      valueencoding  = (normal / escaped)*
      normal         = ~r"[^\\0()*\\\\]"
      escaped        = ESC HEX HEX
      HEX            = DIGIT / ~r"[A-Fa-f]"

      EQUALS   = "="
      TILDE    = "~"
      LANGLE   = "<"
      RANGLE   = ">"
      COLON    = ":"
      ASTERISK = "*"
      DOT      = "."
      HYPHEN   = "-"
      SEMI     = ";"
      ESC      = "\\\\"
'''

rfc4515_filter_grammar = '''
      standard_filter = LPAREN filtercomp RPAREN
      filtercomp      = and / or / not / rfc4515_ava
      and             = AMPERSAND filterlist
      or              = VERTBAR filterlist
      not             = EXCLAMATION standard_filter
      filterlist      = standard_filter+
''' + ava_grammar + '''
      LPAREN      = "("
      RPAREN      = ")"
      AMPERSAND   = "&"
      VERTBAR     = "|"
      EXCLAMATION = "!"
'''

_rfc4515_filter_grammar = Grammar(rfc4515_filter_grammar)


def parse_standard_filter(filter_str):
    """Parse an RFC 4515 filter string to an rfc4511.Filter"""

    try:
        filter_node = _rfc4515_filter_grammar.parse(filter_str)
        return _handle_standard_filter(filter_node)
    except ParseError as e:
        raise LDAPError(str(e))


def _handle_standard_filter(filter_node):
    fil = rfc4511.Filter()
    filtercomp = filter_node.children[1]
    for child in filtercomp.children:
        if child.expr_name == 'and':
            filterlist = child.children[1]
            and_set = rfc4511.And()
            for i, node in enumerate(filterlist.children):
                and_set.setComponentByPosition(i, _handle_standard_filter(node))
            fil.setComponentByName('and', and_set)
        elif child.expr_name == 'or':
            filterlist = child.children[1]
            or_set = rfc4511.Or()
            for i, node in enumerate(filterlist.children):
                or_set.setComponentByPosition(i, _handle_standard_filter(node))
            fil.setComponentByName('or', or_set)
        elif child.expr_name == 'not':
            node = child.children[1]
            not_filter = rfc4511.Not()
            not_filter.setComponentByName('innerNotFilter', _handle_standard_filter(node))
            fil.setComponentByName('not', not_filter)
        elif child.expr_name == 'rfc4515_ava':
            _handle_rfc4515_ava(fil, child)
        else:
            raise LDAPError('Unhandled condition while parsing filter')
    return fil


def _handle_rfc4515_ava(fil, ava_node):
    ava_type = ava_node.children[0]
    if ava_type.expr_name == 'simple':
        attr_type = ava_type.children[0].text
        filtertype = ava_type.children[1].children[0].expr_name
        attr_value = ava_type.children[2].text
        if filtertype == "EQUALS":
            component = 'equalityMatch'
            ava = rfc4511.EqualityMatch()
        elif filtertype == "approx":
            component = 'approxMatch'
            ava = rfc4511.ApproxMatch()
        elif filtertype == "greaterorequal":
            component = 'greaterOrEqual'
            ava = rfc4511.GreaterOrEqual()
        elif filtertype == "lessorequal":
            component = 'lessOrEqual'
            ava = rfc4511.LessOrEqual()
        else:
            raise LDAPError('Unhandled condition while parsing filter')
        ava.setComponentByName('attributeDesc', rfc4511.AttributeDescription(attr_type))
        ava.setComponentByName('assertionValue', rfc4511.AssertionValue(attr_value))
        fil.setComponentByName(component, ava)
    elif ava_type.expr_name == 'substring':
        attr_type = ava_type.children[0].text

        # detect the special case that this should be rfc4511.Present
        if ava_type.children[2].text == '' and ava_type.children[3].text == '*' and ava_type.children[4].text == '':
            fil.setComponentByName('present', rfc4511.Present(attr_type))

        # standard substring
        else:
            subf = rfc4511.SubstringFilter()
            subf.setComponentByName('type', rfc4511.AttributeDescription(attr_type))
            subs = rfc4511.Substrings()
            i = 0
            if ava_type.children[2].text != '':
                c = rfc4511.Substring()
                c.setComponentByName('initial', rfc4511.Initial(ava_type.children[2].text))
                subs.setComponentByPosition(i, c)
                i += 1
            if ava_type.children[3].text != '*':
                for any_sub in ava_type.children[3].children[1].children:
                    c = rfc4511.Substring()
                    c.setComponentByName('any', rfc4511.Any(any_sub.children[0].text))
                    subs.setComponentByPosition(i, c)
                    i += 1
            if ava_type.children[4].text != '':
                c = rfc4511.Substring()
                c.setComponentByName('final', rfc4511.Final(ava_type.children[4].text))
                subs.setComponentByPosition(i, c)
            subf.setComponentByName('substrings', subs)
            fil.setComponentByName('substrings', subf)
    elif ava_type.expr_name == 'extensible':
        ext_filter = ava_type.children[0]

        num_children = len(ext_filter.children)
        if num_children == 6:
            attr = ext_filter.children[0].text
            dnattrs = (ext_filter.children[1].text == ':dn')
            rule = ext_filter.children[2].text[1:]
            val = ext_filter.children[5].text
        elif num_children == 5:
            attr = None
            dnattrs = (ext_filter.children[0].text == ':dn')
            rule = ext_filter.children[1].text[1:]
            val = ext_filter.children[4].text
        else:
            raise LDAPError('Unhandled condition while parsing filter')

        xm = rfc4511.ExtensibleMatch()
        xm.setComponentByName('matchValue', rfc4511.MatchValue(val))
        xm.setComponentByName('dnAttributes', rfc4511.DnAttributes(dnattrs))
        if attr:
            xm.setComponentByName('type', rfc4511.Type(attr))
        if rule:
            xm.setComponentByName('matchingRule', rfc4511.MatchingRule(rule))
        fil.setComponentByName('extensibleMatch', xm)
    else:
        raise LDAPError('Unhandled condition while parsing filter')


laurelin_logic_grammar = '''
    filter      = component or_exp*
    component   = term and_exp*
    or_exp      = SPACE OR SPACE component
    and_exp     = SPACE AND SPACE term
    not_exp     = NOT SPACE term
    paren_term  = "(" SPACE filter SPACE ")"
    SPACE       = ~"[ \t]*"
    OR          = "OR"
    AND         = "AND"
    NOT         = "NOT"
'''

laurelin_filter_grammar = laurelin_logic_grammar + '''
    term        = not_exp / paren_term / ava
    ava         = "(" rfc4515_ava ")"
''' + ava_grammar

_laurelin_filter_grammar = Grammar(laurelin_filter_grammar)


def parse_simple_filter(simple_filter_str):
    """Laurelin defines its own, simpler format for filter strings. It uses the
    RFC 4515 standard format for the various comparison expressions, but with
    SQL-style logic operations. (Fully standard RFC 4515 filters are fully
    supported and used by default)
    """

    try:
        filter_node = _laurelin_filter_grammar.parse(simple_filter_str)
        return _handle_simple_filter(filter_node)
    except ParseError as e:
        raise LDAPError(str(e))


def _handle_simple_filter(filter_node):
    first_component = _handle_component(filter_node.children[0])

    if filter_node.children[1].text != '':
        # got an OR
        fil = rfc4511.Filter()
        or_set = rfc4511.Or()
        i = 0
        or_set.setComponentByPosition(i, first_component)
        i += 1
        or_exps = filter_node.children[1].children
        for or_exp in or_exps:
            or_component = or_exp.children[3]
            or_set.setComponentByPosition(i, _handle_component(or_component))
            i += 1
        fil.setComponentByName('or', or_set)
    else:
        # got a single component
        fil = first_component
    return fil


def _handle_component(component_node):
    term_node = component_node.children[0]
    first_term = _handle_term(term_node)
    if component_node.children[1].text != '':
        # got AND terms
        fil = rfc4511.Filter()
        and_set = rfc4511.And()
        i = 0
        and_set.setComponentByPosition(i, first_term)
        i += 1
        for and_exp in component_node.children[1].children:
            and_set.setComponentByPosition(i, _handle_term(and_exp.children[3]))
            i += 1
        fil.setComponentByName('and', and_set)
    else:
        # got a single term
        fil = first_term
    return fil


def _handle_term(term_node):
    term_type = term_node.children[0]
    if term_type.expr_name == 'not_exp':
        fil = rfc4511.Filter()
        not_filter = rfc4511.Not()
        not_filter.setComponentByName('innerNotFilter', _handle_term(term_type.children[2]))
        fil.setComponentByName('not', not_filter)
    elif term_type.expr_name == 'paren_term':
        fil = _handle_simple_filter(term_type.children[2])

    # support both simple and unified grammar
    elif term_type.expr_name == 'ava':
        fil = rfc4511.Filter()
        _handle_rfc4515_ava(fil, term_type.children[1])
    elif term_type.expr_name == 'standard_filter':
        fil = _handle_standard_filter(term_type)

    else:
        raise LDAPError('Unhandled condition while parsing filter')
    return fil


unified_filter_grammar = laurelin_logic_grammar + '''
    term        = not_exp / paren_term / standard_filter
''' + rfc4515_filter_grammar

_unified_filter_grammar = Grammar(unified_filter_grammar)


def parse(filter_str):
    """Parse unified filter syntax. Fully compatible with standard filters, laurelin simple filters, and any
    intermingling of the two.
    """

    try:
        filter_node = _unified_filter_grammar.parse(filter_str)
        return _handle_simple_filter(filter_node)
    except ParseError as e:
        raise LDAPError(str(e))


def rfc4511_filter_to_rfc4515_string(fil):
    """Reverse :func:`parse_standard_filter`, mainly used for testing

    :param Filter fil: An rfc4511.Filter object
    :return: An RFC 4515 compatible filter string
    """
    filter_type = fil.getName()
    if filter_type == 'and':
        and_obj = fil.getComponent()
        ret = '(&{0})'.format(_reverse_filterset(and_obj))
    elif filter_type == 'or':
        or_obj = fil.getComponent()
        ret = '(|{0})'.format(_reverse_filterset(or_obj))
    elif filter_type == 'not':
        not_obj = fil.getComponent()
        not_filter = not_obj.getComponentByName('innerNotFilter')
        ret = '(!{0})'.format(rfc4511_filter_to_rfc4515_string(not_filter))
    elif filter_type == 'equalityMatch':
        ava = fil.getComponent()
        ret = '({0}={1})'.format(six.text_type(ava.getComponentByName('attributeDesc')),
                                 six.text_type(ava.getComponentByName('assertionValue')))
    elif filter_type == 'substrings':
        subs_obj = fil.getComponent()
        attr_type = six.text_type(subs_obj.getComponentByName('type'))
        subs = subs_obj.getComponentByName('substrings')
        n = len(subs)
        sub_name = ''
        sub_strs = []
        first_type = subs.getComponentByPosition(0).getName()
        if first_type != 'initial':
            sub_strs.append('')
        for i in range(n):
            sub_obj = subs.getComponentByPosition(i)
            sub_name = sub_obj.getName()
            sub_str = six.text_type(sub_obj.getComponent())
            sub_strs.append(sub_str)
        if sub_name != 'final' and sub_strs[-1] != '':
            sub_strs.append('')
        ret = '({0}={1})'.format(attr_type, '*'.join(sub_strs))
    elif filter_type == 'greaterOrEqual':
        ava = fil.getComponent()
        ret = '({0}>={1})'.format(six.text_type(ava.getComponentByName('attributeDesc')),
                                  six.text_type(ava.getComponentByName('assertionValue')))
    elif filter_type == 'lessOrEqual':
        ava = fil.getComponent()
        ret = '({0}<={1})'.format(six.text_type(ava.getComponentByName('attributeDesc')),
                                  six.text_type(ava.getComponentByName('assertionValue')))
    elif filter_type == 'present':
        present_obj = fil.getComponent()
        attr_type = six.text_type(present_obj)
        ret = '({0}=*)'.format(attr_type)
    elif filter_type == 'approxMatch':
        ava = fil.getComponent()
        ret = '({0}~={1})'.format(six.text_type(ava.getComponentByName('attributeDesc')),
                                  six.text_type(ava.getComponentByName('assertionValue')))
    elif filter_type == 'extensibleMatch':
        xm_obj = fil.getComponent()

        rule = ''
        rule_obj = xm_obj.getComponentByName('matchingRule')
        if rule_obj.isValue:
            rule = ':' + six.text_type(rule_obj)

        attr = ''
        attr_obj = xm_obj.getComponentByName('type')
        if attr_obj.isValue:
            attr = six.text_type(attr_obj)

        dn_attrs = ':dn' if bool(xm_obj.getComponentByName('dnAttributes')) else ''

        value = six.text_type(xm_obj.getComponentByName('matchValue'))

        ret = '({0}{1}{2}:={3})'.format(attr, dn_attrs, rule, value)
    else:
        raise LDAPError('Unhandled condition while constructing filter string')
    return ret


def _reverse_filterset(filterset):
    n = len(filterset)
    ret = ''
    for i in range(n):
        ret += rfc4511_filter_to_rfc4515_string(filterset.getComponentByPosition(i))
    return ret



from __future__ import absolute_import
from . import user_base
from .user_base import BaseLaurelinExtension
from ..exceptions import LDAPExtensionError
from importlib import import_module
import logging

logger = logging.getLogger('laurelin.ldap.extensible')

# extensions must define a class with this name
EXTENSION_CLSNAME = 'LaurelinExtension'

# extensions may define classes following this pattern, where {0} is one of "LDAP" or "LDAPObject"
CLASS_EXTENSION_FMT = 'Laurelin{0}Extension'


def add_extension(modname):
    """Import an extension and prepare it for binding under its internally-defined name to LDAP and/or LDAPObject
    depending which extension classes are defined. This is only needed for extensions not yet patched into
    AVAILABLE_EXTENSIONS.

    :param str modname: The string module name containing an extension, can be any importable module, e.g.
                        "laurelin.extensions.netgroups"
    :rtype: None
    """

    # import the extension and get the extension class
    mod = _import_extension(modname)
    ext_attr_name = getattr(mod, EXTENSION_CLSNAME).NAME

    # check if the extension has already been added, return if this exact module is already loaded, exception if its
    # a different module
    try:
        existing_ext_mod = Extensible.ADDITIONAL_EXTENSIONS[ext_attr_name]
        if existing_ext_mod is not mod:
            # the 2 extension modules are not identical; report that a different extension by this NAME is already
            # added
            raise LDAPExtensionError('NAME {0} is already loaded as an additional extension {1}'.format(
                ext_attr_name, existing_ext_mod.__name__
            ))
        else:
            # the 2 extension modules are identical; quietly move on
            logger.debug('Extension module {0} with NAME {1} has already been added'.format(modname, ext_attr_name))
            return
    except KeyError:
        # no existing additional extension with the same NAME
        pass

    # check if the extension is predefined, return if the predefined extension uses the same modname, exception
    # otherwise
    try:
        ext_info = Extensible.AVAILABLE_EXTENSIONS[ext_attr_name]
        if ext_info['module'] != modname:
            # the module names of the 2 extensions differ; assume they are different extensions and report that the
            # NAME is already in use
            raise LDAPExtensionError('{0} is already defined as an extension name'.format(ext_attr_name))
        else:
            # the modules names of the 2 extensions are equal; assume they are the same extension and quietly move on
            logger.debug('Extension module {0} does not need to be added'.format(modname))
            return
    except KeyError:
        # no predefined extension by this NAME
        pass

    # store the imported module for later use
    Extensible.ADDITIONAL_EXTENSIONS[ext_attr_name] = mod


def _import_extension(modname):
    """Import an extension module and run setup functions if necessary"""
    mod = import_module(modname)
    flag_attr = '__LAURELIN_EXTENSION_SETUP_COMPLETE'
    if not getattr(mod, flag_attr, False):
        # need to setup extension
        logger.info('Setting up extension {0}'.format(modname))
        try:
            ext_cls = getattr(mod, EXTENSION_CLSNAME)
            if not issubclass(ext_cls, BaseLaurelinExtension):
                raise LDAPExtensionError('{0}.{1} does not subclass laurelin.ldap.BaseLaurelinExtension'.format(
                    modname, EXTENSION_CLSNAME
                ))

            # check if the class defined a NAME
            if ext_cls.NAME == BaseLaurelinExtension.NAME:
                raise LDAPExtensionError('Extension {0}.{1} does not define a NAME'.format(modname, EXTENSION_CLSNAME))

            # call one-time setup functions
            ext_obj = ext_cls()
            ext_obj.require()

            # store the instance on a class attribute
            ext_cls.INSTANCE = ext_obj
        except AttributeError:
            raise LDAPExtensionError('Extension {0} must define a class {1}'.format(modname, EXTENSION_CLSNAME))
        setattr(mod, flag_attr, True)
    return mod


class Extensible(object):
    """Base for automatically-generated extension property classes"""

    AVAILABLE_EXTENSIONS = {
        'base_schema': {
            'module': 'laurelin.extensions.base_schema',
            'pip_package': None,  # built-in
            'docstring': 'The standard base schema from various RFCs'
        },
        'descattrs': {
            'module': 'laurelin.extensions.descattrs',
            'pip_package': None,  # built-in
            'docstring': 'The built-in description attributes extension',
        },
        'netgroups': {
            'module': 'laurelin.extensions.netgroups',
            'pip_package': None,  # built-in
            'docstring': 'The built-in NIS netgroups extension',
        },
        'paged_results': {
            'module': 'laurelin.extensions.pagedresults',
            'pip_package': None,  # built-in
            'docstring': 'Built-in extension defining standard paged results control for search'
        },
    }

    ADDITIONAL_EXTENSIONS = {}

    def __init__(self):
        self._extension_instances = {}
        self._built_in_only = False

    def _get_extension_instance(self, name):
        """This gets called and returned by auto-generated @property methods as their only line"""
        if self._built_in_only and self.AVAILABLE_EXTENSIONS[name]['pip_package'] is not None:
            raise RuntimeError('3rd-party extensions have been disabled')
        try:
            return self._extension_instances[name]
        except KeyError:
            pass
        extinfo = self.AVAILABLE_EXTENSIONS[name]
        modname = extinfo['module']
        try:
            mod = _import_extension(modname)
        except ImportError:
            pip_package = extinfo['pip_package']
            if pip_package is None:
                raise LDAPExtensionError('Error importing built-in extension {0}. This may be a bug.'.format(modname))
            else:
                raise LDAPExtensionError('Error importing {0}, the extension may not be installed. Try `pip '
                                         'install {1}`'.format(modname, pip_package))
        obj = self._create_extension_instance(name, mod)
        return obj

    def _create_extension_instance(self, name, mod):
        raise NotImplementedError()

    def __getattr__(self, item):
        # this acts like the auto-generated @property methods but fully dynamic, using modules that have been added
        # using add_extension()
        try:
            return self._extension_instances[item]
        except KeyError:
            pass
        try:
            ext_mod = self.ADDITIONAL_EXTENSIONS[item]
            if self._built_in_only:
                raise RuntimeError('3rd-party extensions have been disabled')
        except KeyError:
            raise AttributeError(item)
        obj = self._create_extension_instance(item, ext_mod)
        return obj


class ExtensibleClass(Extensible):
    """Base for auto-generated classes inherited by LDAP and LDAPObject"""

    EXTENSIBLE_CLASSES = ('LDAP', 'LDAPObject')

    def __init__(self):
        Extensible.__init__(self)
        self._extended_classname = None

    def _get_extended_classname(self):
        """Find the name of the class we are extending. Allows users to subclass LDAP or LDAPObject"""
        if self._extended_classname:
            return self._extended_classname
        else:
            for cls in self.__class__.__mro__:
                if cls.__name__ in self.EXTENSIBLE_CLASSES:
                    self._extended_classname = cls.__name__
                    return cls.__name__
            raise TypeError('This class does not inherit a known extensible class')

    def _create_extension_instance(self, name, mod):
        """Creates a new instance of the class extension for this class defined in mod"""
        my_classname = self._get_extended_classname()
        ext_classname = CLASS_EXTENSION_FMT.format(my_classname)
        base_classname = 'Base' + ext_classname
        try:
            cls = getattr(mod, ext_classname)
        except AttributeError:
            raise LDAPExtensionError('Extension {0} does not define an extension class for {1}'.format(
                mod.__name__, my_classname
            ))
        if not issubclass(cls, getattr(user_base, base_classname)):
            raise LDAPExtensionError('Extension class {0}.{1} does not subclass laurelin.ldap.{2}'.format(
                mod.__name__, ext_classname, base_classname
            ))
        obj = cls(parent=self)
        self._extension_instances[name] = obj
        return obj


class ExtensionsBase(Extensible):
    """Base for the auto-generated class giving access to all LaurelinExtension instances"""
    def _create_extension_instance(self, name, mod):
        instance = getattr(mod, EXTENSION_CLSNAME).INSTANCE
        return self._extension_instances.setdefault(name, instance)



from __future__ import absolute_import
from .constants import DELETE_ALL
from .utils import CaseIgnoreDict
import six


class AttrsDict(CaseIgnoreDict):
    """Stores attributes and provides utility methods without any server or object affinity

     Dict keys are case-insensitive attribute names, and dict values are a list of attribute values
    """

    def get_attr(self, attr):
        """Get an attribute's values, or an empty list if the attribute is not defined

        :param attr: The name of the attribute
        :return: A list of values
        :rtype: list
        """
        return self.get(attr, [])

    def iterattrs(self):
        """Iterate all attributes of this object. Yields ``(attr, value)`` tuples."""
        for attr, vals in six.iteritems(self):
            for val in vals:
                yield (attr, val)

    def deepcopy(self):
        """Return a native dict copy of self."""
        ret = {}
        for attr, vals in six.iteritems(self):
            ret[attr] = []
            for val in vals:
                ret[attr].append(val)
        return ret

    ## dict overrides for case-insensitive keys and enforcing types

    def __init__(self, attrs_dict=None):
        CaseIgnoreDict.__init__(self, attrs_dict)

    def __contains__(self, attr):
        try:
            return len(self[attr]) > 0
        except KeyError:
            return False

    def __setitem__(self, attr, values):
        AttrsDict.validate_attr(attr)
        AttrsDict.validate_values(values)
        CaseIgnoreDict.__setitem__(self, attr, values)

    def setdefault(self, attr, default=None):
        AttrsDict.validate_attr(attr)
        if default is None:
            default = []
        try:
            AttrsDict.validate_values(default)
        except TypeError as e:
            raise TypeError('invalid default - {0}'.format(str(e)))
        return CaseIgnoreDict.setdefault(self, attr, default)

    def update(self, attrs_dict):
        AttrsDict.validate(attrs_dict)
        CaseIgnoreDict.update(self, attrs_dict)

    @staticmethod
    def validate(attrs_dict):
        """Validate that ``attrs_dict`` is either already an :class:`.AttrsDict` or that it conforms to the required
        ``dict(str, list[str or bytes])`` typing.

        :param dict attrs_dict: The dictionary to validate for use as an attributes dictionary
        :rtype: None
        :raises TypeError: when the dict is invalid
        """
        if isinstance(attrs_dict, AttrsDict):
            return
        if not isinstance(attrs_dict, dict):
            raise TypeError('must be dict')
        for attr in attrs_dict:
            AttrsDict.validate_attr(attr)
            AttrsDict.validate_values(attrs_dict[attr])

    @staticmethod
    def validate_attr(attr):
        """Validate that ``attr`` is a valid attribute name.

        :param str attr: The string to validate for use as an attribute name
        :rtype: None
        :raises TypeError: when the string is invalid
        """
        if not isinstance(attr, six.string_types):
            raise TypeError('attribute name must be string')

    @staticmethod
    def validate_values(attr_val_list):
        """Validate that ``attr_val_list`` conforms to the required ``list[str or bytes]`` typing. Also allows the
        DELETE_ALL constant.

        :param list attr_val_list: The list to validate for use as an attribute value list.
        :rtype: None
        :raises TypeError: when the list is invalid
        """
        if attr_val_list is DELETE_ALL:
            return
        if not isinstance(attr_val_list, list):
            raise TypeError('must be list')
        for val in attr_val_list:
            if not isinstance(val, six.string_types) and not isinstance(val, six.binary_type):
                raise TypeError('attribute values must be string or bytes')



"""RFC 2696 Simple Paged Results Manipulation

This adds a control to support paging results. Use the control keyword ``paged`` with search methods. Returns a cookie
on the ``page_cookie`` response attribute which can be found on the results handle after all paged results have been
received. See example below.

Note: Do not use this extension to simply limit the total number of results. The search methods accept a ``limit``
keyword out of the box for this purpose.

Example usage::

    from laurelin.ldap import LDAP
    LDAP.activate_extension('laurelin.extensions.pagedresults')

    with LDAP() as ldap:
        search = ldap.base.search(paged=10)
        page1_results = list(search)

        search = ldap.base.search(paged=(10, search.page_cookie))
        page2_results = list(search)

        # ...

        if not search.page_cookie:
            print('Got all pages')

Note: When getting pages in a loop, you may set the cookie value to an empty string on the first iteration, e.g.::

    ldap.base.search(paged=(10, ''))
"""

from laurelin.ldap import Control, BaseLaurelinExtension, LDAPError
from laurelin.ldap.protoutils import get_string_component
from laurelin.ldap.rfc4511 import Integer0ToMax
from laurelin.ldap.pyasn1.codec.ber.encoder import encode as ber_encode
from laurelin.ldap.pyasn1.codec.ber.decoder import decode as ber_decode
from laurelin.ldap.pyasn1.type.univ import OctetString, Sequence
from laurelin.ldap.pyasn1.type.namedtype import NamedTypes, NamedType

OID = '1.2.840.113556.1.4.319'


class Size(Integer0ToMax):
    pass


class Cookie(OctetString):
    pass


class RealSearchControlValue(Sequence):
    # realSearchControlValue ::= SEQUENCE {
    #         size            INTEGER (0..maxInt),
    #                                 -- requested page size from client
    #                                 -- result set size estimate from server
    #         cookie          OCTET STRING
    # }
    componentType = NamedTypes(NamedType('size', Size()),
                               NamedType('cookie', Cookie()))


class LaurelinExtension(BaseLaurelinExtension):
    NAME = 'paged_results'

    OID = OID

    class PagedResultsControl(Control):
        REQUEST_OID = OID
        RESPONSE_OID = OID
        method = ('search',)
        keyword = 'paged'
        response_attr = 'page_cookie'

        def prepare(self, ctrl_value, criticality):
            """Prepare the paged results control value

            :param ctrl_value: Either an integer page size or a tuple of (page size, cookie)
            :type ctrl_value: int or tuple
            :param criticality: True if the control is critical, false otherwise
            :return: The protocol-level control object
            """
            if isinstance(ctrl_value, int):
                page_size = ctrl_value
                cookie = ''
            elif isinstance(ctrl_value, tuple):
                page_size, cookie = ctrl_value
            else:
                raise TypeError('Must be int or tuple')
            real_ctrl_value = RealSearchControlValue()
            real_ctrl_value.setComponentByName('size', Size(page_size))
            real_ctrl_value.setComponentByName('cookie', Cookie(cookie))
            real_ctrl_value = ber_encode(real_ctrl_value)
            return Control.prepare(self, real_ctrl_value, criticality)

        def handle(self, ctrl_value):
            real_ctrl_value, raw = ber_decode(ctrl_value, asn1Spec=RealSearchControlValue())
            if raw:
                raise LDAPError('Unexpected leftover bits in response control value')
            cookie = get_string_component(real_ctrl_value, 'cookie')
            return cookie
